# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral.db.v2 import api as db_api
from mistral.lang import parser as spec_parser
from mistral.services import workbooks as wb_service
from mistral.services import workflows as wf_service
from mistral.tests.unit import base
from mistral.tests.unit.engine import base as engine_base
from mistral.workflow import states


class SpecificationCachingTest(base.DbTestCase):
    def test_workflow_spec_caching(self):
        wf_text = """
        version: '2.0'

        wf:
          tasks:
            task1:
              action: std.echo output="Echo"
        """

        wfs = wf_service.create_workflows(wf_text)

        self.assertEqual(0, spec_parser.get_wf_execution_spec_cache_size())
        self.assertEqual(0, spec_parser.get_wf_definition_spec_cache_size())

        wf_spec = spec_parser.get_workflow_spec_by_definition_id(
            wfs[0].id,
            wfs[0].updated_at
        )

        self.assertIsNotNone(wf_spec)
        self.assertEqual(0, spec_parser.get_wf_execution_spec_cache_size())
        self.assertEqual(1, spec_parser.get_wf_definition_spec_cache_size())

    def test_workflow_spec_cache_update_via_workflow_service(self):
        wf_text = """
        version: '2.0'

        wf:
          tasks:
            task1:
              action: std.echo output="Echo"
        """

        wfs = wf_service.create_workflows(wf_text)

        self.assertEqual(0, spec_parser.get_wf_execution_spec_cache_size())
        self.assertEqual(0, spec_parser.get_wf_definition_spec_cache_size())

        wf_spec = spec_parser.get_workflow_spec_by_definition_id(
            wfs[0].id,
            wfs[0].updated_at
        )

        self.assertEqual(1, len(wf_spec.get_tasks()))
        self.assertEqual(0, spec_parser.get_wf_execution_spec_cache_size())
        self.assertEqual(1, spec_parser.get_wf_definition_spec_cache_size())

        # Now update workflow definition and check that cache is updated too.

        wf_text = """
        version: '2.0'

        wf:
          tasks:
            task1:
              action: std.echo output="1"

            task2:
              action: std.echo output="2"
        """

        wfs = wf_service.update_workflows(wf_text)

        self.assertEqual(1, spec_parser.get_wf_definition_spec_cache_size())

        wf_spec = spec_parser.get_workflow_spec_by_definition_id(
            wfs[0].id,
            wfs[0].updated_at
        )

        self.assertEqual(2, len(wf_spec.get_tasks()))
        self.assertEqual(2, spec_parser.get_wf_definition_spec_cache_size())
        self.assertEqual(0, spec_parser.get_wf_execution_spec_cache_size())

    def test_workflow_spec_cache_update_via_workbook_service(self):
        wb_text = """
        version: '2.0'

        name: wb

        workflows:
          wf:
            tasks:
              task1:
                action: std.echo output="Echo"
        """

        wb_service.create_workbook_v2(wb_text)

        self.assertEqual(0, spec_parser.get_wf_execution_spec_cache_size())
        self.assertEqual(0, spec_parser.get_wf_definition_spec_cache_size())

        wf = db_api.get_workflow_definition('wb.wf')

        wf_spec = spec_parser.get_workflow_spec_by_definition_id(
            wf.id,
            wf.updated_at
        )

        self.assertEqual(1, len(wf_spec.get_tasks()))
        self.assertEqual(0, spec_parser.get_wf_execution_spec_cache_size())
        self.assertEqual(1, spec_parser.get_wf_definition_spec_cache_size())

        # Now update workflow definition and check that cache is updated too.

        wb_text = """
        version: '2.0'

        name: wb

        workflows:
          wf:
            tasks:
              task1:
                action: std.echo output="1"

              task2:
                action: std.echo output="2"
        """

        wb_service.update_workbook_v2(wb_text)

        self.assertEqual(0, spec_parser.get_wf_execution_spec_cache_size())
        self.assertEqual(1, spec_parser.get_wf_definition_spec_cache_size())

        wf = db_api.get_workflow_definition(wf.id)

        wf_spec = spec_parser.get_workflow_spec_by_definition_id(
            wf.id,
            wf.updated_at
        )

        self.assertEqual(2, len(wf_spec.get_tasks()))
        self.assertEqual(0, spec_parser.get_wf_execution_spec_cache_size())
        self.assertEqual(2, spec_parser.get_wf_definition_spec_cache_size())

    def test_cache_workflow_spec_by_execution_id(self):
        wf_text = """
        version: '2.0'

        wf:
          tasks:
            task1:
              action: std.echo output="Echo"
        """

        wfs = wf_service.create_workflows(wf_text)

        self.assertEqual(0, spec_parser.get_wf_execution_spec_cache_size())
        self.assertEqual(0, spec_parser.get_wf_definition_spec_cache_size())

        wf_def = wfs[0]

        wf_spec = spec_parser.get_workflow_spec_by_definition_id(
            wf_def.id,
            wf_def.updated_at
        )

        self.assertEqual(1, len(wf_spec.get_tasks()))
        self.assertEqual(0, spec_parser.get_wf_execution_spec_cache_size())
        self.assertEqual(1, spec_parser.get_wf_definition_spec_cache_size())

        with db_api.transaction():
            wf_ex = db_api.create_workflow_execution({
                'id': '1-2-3-4',
                'name': 'wf',
                'workflow_id': wf_def.id,
                'spec': wf_spec.to_dict(),
                'state': states.RUNNING
            })

            # Check that we can get a valid spec by execution id.
            wf_spec_by_exec_id = spec_parser.get_workflow_spec_by_execution_id(
                wf_ex.id
            )

        self.assertEqual(1, len(wf_spec_by_exec_id.get_tasks()))

        # Now update workflow definition and check that cache is updated too.

        wf_text = """
        version: '2.0'

        wf:
          tasks:
            task1:
              action: std.echo output="1"

            task2:
              action: std.echo output="2"
        """

        wfs = wf_service.update_workflows(wf_text)

        self.assertEqual(1, spec_parser.get_wf_definition_spec_cache_size())

        wf_spec = spec_parser.get_workflow_spec_by_definition_id(
            wfs[0].id,
            wfs[0].updated_at
        )

        self.assertEqual(2, len(wf_spec.get_tasks()))
        self.assertEqual(2, spec_parser.get_wf_definition_spec_cache_size())
        self.assertEqual(1, spec_parser.get_wf_execution_spec_cache_size())

        # Now finally update execution cache and check that we can
        # get a valid spec by execution id.
        spec_parser.cache_workflow_spec_by_execution_id(wf_ex.id, wf_spec)

        wf_spec_by_exec_id = spec_parser.get_workflow_spec_by_execution_id(
            wf_ex.id
        )

        self.assertEqual(2, len(wf_spec_by_exec_id.get_tasks()))


class SpecificationCachingEngineTest(engine_base.EngineTestCase):
    def test_cache_workflow_spec_no_duplicates(self):
        wfs_text = """
        version: '2.0'

        wf:
          tasks:
            task1:
              action: std.noop
              on-success:
                - task2
                - task3

            task2:
              workflow: sub_wf my_param="val1"

            task3:
              workflow: sub_wf my_param="val2"

        sub_wf:
          input:
            - my_param

          tasks:
            task1:
              action: std.echo output="Param value is <% $.my_param %>"
        """

        wfs = wf_service.create_workflows(wfs_text)

        self.assertEqual(2, len(wfs))

        self.assertEqual(0, spec_parser.get_wf_execution_spec_cache_size())
        self.assertEqual(0, spec_parser.get_wf_definition_spec_cache_size())

        wf_ex = self.engine.start_workflow('wf')

        self.await_workflow_success(wf_ex.id)

        # We expect to have a cache entry for every workflow execution
        # but two of them should refer to the same object.
        self.assertEqual(3, spec_parser.get_wf_execution_spec_cache_size())
        self.assertEqual(2, spec_parser.get_wf_definition_spec_cache_size())

        sub_wf_execs = db_api.get_workflow_executions(name='sub_wf')

        self.assertEqual(2, len(sub_wf_execs))

        spec1 = spec_parser.get_workflow_spec_by_execution_id(
            sub_wf_execs[0].id
        )
        spec2 = spec_parser.get_workflow_spec_by_execution_id(
            sub_wf_execs[1].id
        )

        self.assertIs(spec1, spec2)



#  Copyright 2019 - Nokia Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import unittest
import uuid
from uuid import UUID


from {{cookiecutter.pkg_name}}.expression_functions import my_function_


class MyTestCase(unittest.TestCase):
    def test_my_function(self):
        context = {'__execution': {'id': uuid.uuid4().hex}}
        result = my_function_(context)

        # should not throw an exception
        UUID(str(result), version=5)



# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_config import cfg

from mistral.db.v2 import api as db_api
from mistral.db.v2.sqlalchemy import models
from mistral import exceptions as exc
from mistral import expressions as expr
from mistral.services import workbooks as wb_service
from mistral.services import workflows as wf_service
from mistral.tests.unit import base as test_base
from mistral.tests.unit.engine import base as engine_test_base
from mistral import utils
from mistral.workflow import data_flow
from mistral.workflow import states

import sys


# Use the set_default method to set value otherwise in certain test cases
# the change in value is not permanent.
cfg.CONF.set_default('auth_enable', False, group='pecan')


class DataFlowEngineTest(engine_test_base.EngineTestCase):
    def test_linear_dataflow(self):
        wf_text = """---
        version: '2.0'

        wf:
          type: direct

          tasks:
            task1:
              action: std.echo output="Hi"
              publish:
                hi: <% task(task1).result %>
              on-success:
                - task2

            task2:
              action: std.echo output="Morpheus"
              publish:
                to: <% task(task2).result %>
              on-success:
                - task3

            task3:
              publish:
                result: "<% $.hi %>, <% $.to %>! Your <% env().from %>."
        """

        wf_service.create_workflows(wf_text)

        # Start workflow.
        wf_ex = self.engine.start_workflow('wf', env={'from': 'Neo'})

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            tasks = wf_ex.task_executions

        self.assertEqual(states.SUCCESS, wf_ex.state)

        task1 = self._assert_single_item(tasks, name='task1')
        task2 = self._assert_single_item(tasks, name='task2')
        task3 = self._assert_single_item(tasks, name='task3')

        self.assertEqual(states.SUCCESS, task3.state)
        self.assertDictEqual({'hi': 'Hi'}, task1.published)
        self.assertDictEqual({'to': 'Morpheus'}, task2.published)
        self.assertDictEqual(
            {'result': 'Hi, Morpheus! Your Neo.'},
            task3.published
        )

        # Make sure that task inbound context doesn't contain workflow
        # execution info.
        self.assertNotIn('__execution', task1.in_context)

    def test_linear_with_branches_dataflow(self):
        wf_text = """---
        version: '2.0'

        wf:
          type: direct

          tasks:
            task1:
              action: std.echo output="Hi"
              publish:
                hi: <% task(task1).result %>
                progress: "completed task1"
              on-success:
                - notify
                - task2

            task2:
              action: std.echo output="Morpheus"
              publish:
                to: <% task(task2).result %>
                progress: "completed task2"
              on-success:
                - notify
                - task3

            task3:
              publish:
                result: "<% $.hi %>, <% $.to %>! Your <% env().from %>."
                progress: "completed task3"
              on-success:
                - notify

            notify:
              action: std.echo output=<% $.progress %>
              publish:
                progress: <% task(notify).result %>
        """

        wf_service.create_workflows(wf_text)

        # Start workflow.
        wf_ex = self.engine.start_workflow('wf', env={'from': 'Neo'})

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            tasks = wf_ex.task_executions

        self.assertEqual(states.SUCCESS, wf_ex.state)

        task1 = self._assert_single_item(tasks, name='task1')
        task2 = self._assert_single_item(tasks, name='task2')
        task3 = self._assert_single_item(tasks, name='task3')

        notify_tasks = self._assert_multiple_items(tasks, 3, name='notify')
        notify_published_arr = [t.published['progress'] for t in notify_tasks]

        self.assertEqual(states.SUCCESS, task3.state)

        exp_published_arr = [
            {
                'hi': 'Hi',
                'progress': 'completed task1'
            },
            {
                'to': 'Morpheus',
                'progress': 'completed task2'
            },
            {
                'result': 'Hi, Morpheus! Your Neo.',
                'progress': 'completed task3'
            }
        ]

        self.assertDictEqual(exp_published_arr[0], task1.published)
        self.assertDictEqual(exp_published_arr[1], task2.published)
        self.assertDictEqual(exp_published_arr[2], task3.published)

        self.assertIn(
            exp_published_arr[0]['progress'],
            notify_published_arr
        )
        self.assertIn(
            exp_published_arr[1]['progress'],
            notify_published_arr
        )
        self.assertIn(
            exp_published_arr[2]['progress'],
            notify_published_arr
        )

    def test_parallel_tasks(self):
        wf_text = """---
        version: '2.0'

        wf:
          type: direct

          tasks:
            task1:
              action: std.echo output=1
              publish:
                var1: <% task(task1).result %>

            task2:
              action: std.echo output=2
              publish:
                var2: <% task(task2).result %>
        """

        wf_service.create_workflows(wf_text)

        # Start workflow.
        wf_ex = self.engine.start_workflow('wf',)

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            wf_output = wf_ex.output
            tasks = wf_ex.task_executions

        self.assertEqual(states.SUCCESS, wf_ex.state)

        self.assertEqual(2, len(tasks))

        task1 = self._assert_single_item(tasks, name='task1')
        task2 = self._assert_single_item(tasks, name='task2')

        self.assertEqual(states.SUCCESS, task1.state)
        self.assertEqual(states.SUCCESS, task2.state)

        self.assertDictEqual({'var1': 1}, task1.published)
        self.assertDictEqual({'var2': 2}, task2.published)

        self.assertEqual(1, wf_output['var1'])
        self.assertEqual(2, wf_output['var2'])

    def test_parallel_tasks_complex(self):
        wf_text = """---
        version: '2.0'

        wf:
          type: direct

          tasks:
            task1:
              action: std.noop
              publish:
                var1: 1
              on-complete:
                - task12

            task12:
              action: std.noop
              publish:
                var12: 12
              on-complete:
                - task13
                - task14

            task13:
              action: std.fail
              description: |
                Since this task fails we expect that 'var13' won't go into
                context. Only 'var14'.
              publish:
                var13: 13
              on-error:
                - noop

            task14:
              publish:
                var14: 14

            task2:
              publish:
                var2: 2
              on-complete:
                - task21

            task21:
              publish:
                var21: 21
        """

        wf_service.create_workflows(wf_text)

        # Start workflow.
        wf_ex = self.engine.start_workflow('wf')

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            wf_output = wf_ex.output
            tasks = wf_ex.task_executions

        self.assertEqual(states.SUCCESS, wf_ex.state)

        self.assertEqual(6, len(tasks))

        task1 = self._assert_single_item(tasks, name='task1')
        task12 = self._assert_single_item(tasks, name='task12')
        task13 = self._assert_single_item(tasks, name='task13')
        task14 = self._assert_single_item(tasks, name='task14')
        task2 = self._assert_single_item(tasks, name='task2')
        task21 = self._assert_single_item(tasks, name='task21')

        self.assertEqual(states.SUCCESS, task1.state)
        self.assertEqual(states.SUCCESS, task12.state)
        self.assertEqual(states.ERROR, task13.state)
        self.assertEqual(states.SUCCESS, task14.state)
        self.assertEqual(states.SUCCESS, task2.state)
        self.assertEqual(states.SUCCESS, task21.state)

        self.assertDictEqual({'var1': 1}, task1.published)
        self.assertDictEqual({'var12': 12}, task12.published)
        self.assertDictEqual({'var14': 14}, task14.published)
        self.assertDictEqual({'var2': 2}, task2.published)
        self.assertDictEqual({'var21': 21}, task21.published)

        self.assertEqual(1, wf_output['var1'])
        self.assertEqual(12, wf_output['var12'])
        self.assertNotIn('var13', wf_output)
        self.assertEqual(14, wf_output['var14'])
        self.assertEqual(2, wf_output['var2'])
        self.assertEqual(21, wf_output['var21'])

    def test_sequential_tasks_publishing_same_var(self):
        wf_text = """---
        version: '2.0'

        wf:
          type: direct

          tasks:
            task1:
              action: std.echo output="Hi"
              publish:
                greeting: <% task(task1).result %>
              on-success:
                - task2

            task2:
              action: std.echo output="Yo"
              publish:
                greeting: <% task(task2).result %>
              on-success:
                - task3

            task3:
              action: std.echo output="Morpheus"
              publish:
                to: <% task(task3).result %>
              on-success:
                - task4

            task4:
              publish:
                result: "<% $.greeting %>, <% $.to %>! <% env().from %>."
        """

        wf_service.create_workflows(wf_text)

        # Start workflow.
        wf_ex = self.engine.start_workflow('wf', env={'from': 'Neo'})

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            tasks = wf_ex.task_executions

        self.assertEqual(states.SUCCESS, wf_ex.state)

        task1 = self._assert_single_item(tasks, name='task1')
        task2 = self._assert_single_item(tasks, name='task2')
        task3 = self._assert_single_item(tasks, name='task3')
        task4 = self._assert_single_item(tasks, name='task4')

        self.assertEqual(states.SUCCESS, task4.state)
        self.assertDictEqual({'greeting': 'Hi'}, task1.published)
        self.assertDictEqual({'greeting': 'Yo'}, task2.published)
        self.assertDictEqual({'to': 'Morpheus'}, task3.published)
        self.assertDictEqual(
            {'result': 'Yo, Morpheus! Neo.'},
            task4.published
        )

    def test_sequential_tasks_publishing_same_structured(self):
        wf_text = """---
        version: '2.0'

        wf:
          type: direct

          tasks:
            task1:
              publish:
                greeting: {"a": "b"}
              on-success:
                - task2

            task2:
              publish:
                greeting: {}
              on-success:
                - task3

            task3:
              publish:
                result: <% $.greeting %>
        """

        wf_service.create_workflows(wf_text)

        # Start workflow.
        wf_ex = self.engine.start_workflow('wf', env={'from': 'Neo'})

        self.await_workflow_success(wf_ex.id)

        # Note: We need to reread execution to access related tasks.
        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            tasks = wf_ex.task_executions

        self.assertEqual(states.SUCCESS, wf_ex.state)

        task1 = self._assert_single_item(tasks, name='task1')
        task2 = self._assert_single_item(tasks, name='task2')
        task3 = self._assert_single_item(tasks, name='task3')

        self.assertEqual(states.SUCCESS, task3.state)
        self.assertDictEqual({'greeting': {'a': 'b'}}, task1.published)
        self.assertDictEqual({'greeting': {}}, task2.published)
        self.assertDictEqual({'result': {}}, task3.published)

    def test_linear_dataflow_implicit_publish(self):
        wf_text = """---
        version: '2.0'

        wf:
          type: direct

          tasks:
            task1:
              action: std.echo output="Hi"
              on-success:
                - task21
                - task22

            task21:
              action: std.echo output="Morpheus"
              on-success:
                - task4

            task22:
              action: std.echo output="Neo"
              on-success:
                - task4

            task4:
              join: all
              publish:
                result: >
                  <% task(task1).result %>, <% task(task21).result %>!
                  Your <% task(task22).result %>.
        """
        wf_service.create_workflows(wf_text)

        # Start workflow.
        wf_ex = self.engine.start_workflow('wf')

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            tasks = wf_ex.task_executions

        task4 = self._assert_single_item(tasks, name='task4')

        self.assertDictEqual(
            {'result': 'Hi, Morpheus! Your Neo.\n'},
            task4.published
        )

    def test_destroy_result(self):
        wf_text = """---
        version: '2.0'

        wf:
          type: direct

          tasks:
            task1:
              action: std.echo output=["Hi", "John Doe!"]
              publish:
                hi: <% task(task1).result %>
              keep-result: false

        """
        wf_service.create_workflows(wf_text)

        # Start workflow.
        wf_ex = self.engine.start_workflow('wf')

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            tasks = wf_ex.task_executions

            task1 = self._assert_single_item(tasks, name='task1')

            result = data_flow.get_task_execution_result(task1)

        # Published vars are saved.
        self.assertDictEqual(
            {'hi': ["Hi", "John Doe!"]},
            task1.published
        )

        # But all result is cleared.
        self.assertIsNone(result)

    def test_empty_with_items(self):
        wf_text = """---
        version: "2.0"

        wf1_with_items:
           type: direct

           tasks:
             task1:
               with-items: i in <% list() %>
               action: std.echo output= "Task 1.<% $.i %>"
               publish:
                 result: <% task(task1).result %>
        """
        wf_service.create_workflows(wf_text)

        # Start workflow.
        wf_ex = self.engine.start_workflow('wf1_with_items')

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task1 = self._assert_single_item(
                wf_ex.task_executions,
                name='task1'
            )

            result = data_flow.get_task_execution_result(task1)

        self.assertListEqual([], result)

    def test_publish_on_error(self):
        wf_text = """---
        version: '2.0'

        wf:
          type: direct

          output-on-error:
            out: <% $.hi %>

          tasks:
            task1:
              action: std.fail
              publish-on-error:
                hi: hello_from_error
                err: <% task(task1).result %>
        """

        wf_service.create_workflows(wf_text)

        # Start workflow.
        wf_ex = self.engine.start_workflow('wf')

        self.await_workflow_error(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            wf_output = wf_ex.output
            tasks = wf_ex.task_executions

        self.assertEqual(states.ERROR, wf_ex.state)

        task1 = self._assert_single_item(tasks, name='task1')

        self.assertEqual(states.ERROR, task1.state)
        self.assertEqual('hello_from_error', task1.published['hi'])
        self.assertIn(
            'Fail action expected exception',
            task1.published['err']
        )
        self.assertEqual('hello_from_error', wf_output['out'])
        self.assertIn(
            'Fail action expected exception',
            wf_output['result']
        )

    def test_publish_with_all(self):
        wf_text = """---
        version: '2.0'
        wf:
            tasks:
              main-task:
                publish:
                  res_x1: 111
                on-complete:
                  next: complete-task
                  publish:
                    branch:
                      res_x3: 222
                on-success:
                  next: success-task
                  publish:
                    branch:
                      res_x2: 222

              success-task:
                action: std.noop
                publish:
                  success_x2: <% $.res_x2 %>
                  success_x1: <% $.res_x1 %>

              complete-task:
                action: std.noop
                publish:
                  complete_x2: <% $.res_x3 %>
                  complete_x1: <% $.res_x1 %>
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf')
        self.await_workflow_success(wf_ex.id)
        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            wf_output = wf_ex.output
            tasks = wf_ex.task_executions

            main_task = self._assert_single_item(tasks, name='main-task')
            main_task_published_vars = main_task.get("published")
            expected_main_variables = {'res_x3', 'res_x2', 'res_x1'}
            self.assertEqual(set(main_task_published_vars.keys()),
                             expected_main_variables)

            complete_task = self._assert_single_item(tasks,
                                                     name='complete-task')

            complete_task_published_vars = complete_task.get("published")
            expected_complete_variables = {'complete_x2', 'complete_x1'}
            self.assertEqual(set(complete_task_published_vars.keys()),
                             expected_complete_variables)

            success_task = self._assert_single_item(tasks, name='success-task')
            success_task_published_vars = success_task.get("published")
            expected_success_variables = {'success_x2', 'success_x1'}
            self.assertEqual(set(success_task_published_vars.keys()),
                             expected_success_variables)

            all_expected_published_variables = expected_main_variables.union(
                expected_success_variables,
                expected_complete_variables
            )
            self.assertEqual(set(wf_output), all_expected_published_variables)

    def test_publish_no_success(self):
        wf_text = """---
        version: '2.0'
        wf:
            tasks:
              main-task:
                publish:
                  res_x1: 111
                on-complete:
                  next: complete-task
                  publish:
                    branch:
                      res_x3: 222

              complete-task:
                action: std.noop
                publish:
                  complete_x2: <% $.res_x3 %>
                  complete_x1: <% $.res_x1 %>
        """

        wf_service.create_workflows(wf_text)
        wf_ex = self.engine.start_workflow('wf')
        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            wf_output = wf_ex.output
            tasks = wf_ex.task_executions

            main_task = self._assert_single_item(tasks, name='main-task')
            main_task_published_vars = main_task.get("published")
            expected_main_variables = {'res_x3', 'res_x1'}
            self.assertEqual(set(main_task_published_vars.keys()),
                             expected_main_variables)

            complete_task = self._assert_single_item(tasks,
                                                     name='complete-task')
            complete_task_published_vars = complete_task.get("published")
            expected_complete_variables = {'complete_x2', 'complete_x1'}
            self.assertEqual(set(complete_task_published_vars.keys()),
                             expected_complete_variables)

            all_expected_published_variables = expected_main_variables.union(
                expected_complete_variables)
            self.assertEqual(set(wf_output), all_expected_published_variables)

    def test_publish_no_complete(self):
        wf_text = """---
        version: '2.0'
        wf:
            tasks:
              main-task:
                publish:
                  res_x1: 111
                on-success:
                  next: success-task
                  publish:
                    branch:
                      res_x2: 222

              success-task:
                action: std.noop
                publish:
                  success_x2: <% $.res_x2 %>
                  success_x1: <% $.res_x1 %>
        """

        wf_service.create_workflows(wf_text)
        wf_ex = self.engine.start_workflow('wf')
        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            tasks = wf_ex.task_executions
            wf_output = wf_ex.output

            main_task = self._assert_single_item(tasks, name='main-task')

            main_task_published_vars = main_task.get("published")
            expected_main_variables = {'res_x2', 'res_x1'}
            self.assertEqual(set(main_task_published_vars.keys()),
                             expected_main_variables)

            success_task = self._assert_single_item(tasks, name='success-task')
            success_task_published_vars = success_task.get("published")
            expected_success_variables = {'success_x2', 'success_x1'}
            self.assertEqual(set(success_task_published_vars.keys()),
                             expected_success_variables)

            all_expected_published_variables = expected_main_variables.union(
                expected_success_variables)
            self.assertEqual(set(wf_output), all_expected_published_variables)

    def test_publish_no_regular_publish(self):
        wf_text = """---
        version: '2.0'
        wf2:
            tasks:
              main-task:
                on-success:
                  next: success-task
                  publish:
                    branch:
                      res_x2: 222

              success-task:
                action: std.noop
                publish:
                  success_x2: <% $.res_x2 %>
        """
        wf_service.create_workflows(wf_text)
        wf_ex = self.engine.start_workflow('wf2')
        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            tasks = wf_ex.task_executions
            wf_output = wf_ex.output
            main_task = self._assert_single_item(tasks, name='main-task')
            main_task_published_vars = main_task.get("published")
            expected_main_variables = {'res_x2'}
            self.assertEqual(set(main_task_published_vars.keys()),
                             expected_main_variables)

            success_task = self._assert_single_item(tasks, name='success-task')
            success_task_published_vars = success_task.get("published")
            expected_success_variables = {'success_x2'}
            self.assertEqual(set(success_task_published_vars.keys()),
                             expected_success_variables)

            all_expected_published_variables = expected_main_variables.union(
                expected_success_variables)
            self.assertEqual(set(wf_output), all_expected_published_variables)

    def test_output_on_error_wb_yaql_failed(self):
        wb_text = """---
            version: '2.0'

            name: wb

            workflows:
              wf1:
                type: direct
                output-on-error:
                    message: <% $.message %>

                tasks:
                  task1:
                    workflow: wf2
                    publish-on-error:
                        message: <% task(task1).result.message %>

              wf2:
                type: direct
                output-on-error:
                    message: <% $.not_existing_variable %>

                tasks:
                    task1:
                        action: std.fail
                        publish-on-error:
                            message: <% task(task1).result %>
        """

        wb_service.create_workbook_v2(wb_text)

        # Start workflow.
        wf_ex = self.engine.start_workflow('wb.wf1')

        self.await_workflow_error(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            tasks = wf_ex.task_executions

        self.assertEqual(states.ERROR, wf_ex.state)
        self.assertIn('Failed to evaluate expression in output-on-error!',
                      wf_ex.state_info)
        self.assertIn('$.message', wf_ex.state_info)

        task1 = self._assert_single_item(tasks, name='task1')

        self.assertIn('task(task1).result.message', task1.state_info)

    def test_size_of_output_by_execution_field_size_limit_kb(self):
        wf_text = """
        version: '2.0'

        wf:
          type: direct

          output-on-error:
            custom_error: The action in the task does not exists

          tasks:
            task1:
              action: wrong.task
        """
        # Note: The number 1121 below added as value for field size
        # limit is because the output of workflow error comes as
        # workflow error string + custom error message and total length
        # might be greater than 1121. It varies depending on the length
        # of the custom message. This is a random number value used for
        # test case only.
        cfg.CONF.set_default(
            'execution_field_size_limit_kb',
            1121,
            group='engine'
        )

        kilobytes = cfg.CONF.engine.execution_field_size_limit_kb

        bytes_per_char = sys.getsizeof('s') - sys.getsizeof('')

        total_output_length = int(kilobytes * 1024 / bytes_per_char)

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf', '', None)

        self.await_workflow_error(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            wf_output = wf_ex.output

        self.assertLess(
            len(str(wf_output.get('custom_error'))),
            total_output_length
        )

    def test_override_json_input(self):
        wf_text = """---
        version: 2.0

        wf:
          input:
            - a:
                aa: aa
                bb: bb

          tasks:
            task1:
              action: std.noop
              publish:
                published_a: <% $.a %>
        """

        wf_service.create_workflows(wf_text)

        wf_input = {
            'a': {
                'cc': 'cc',
                'dd': 'dd'
            }
        }

        # Start workflow.
        wf_ex = self.engine.start_workflow('wf', wf_input=wf_input)

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            task1 = wf_ex.task_executions[0]

        self.assertDictEqual(wf_input['a'], task1.published['published_a'])

    def test_branch_publishing_success(self):
        wf_text = """---
        version: 2.0

        wf:
          tasks:
            task1:
              action: std.noop
              on-success:
                publish:
                  branch:
                    my_var: my branch value
                next: task2

            task2:
              action: std.echo output=<% $.my_var %>
        """

        wf_service.create_workflows(wf_text)

        # Start workflow.
        wf_ex = self.engine.start_workflow('wf')

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            tasks = wf_ex.task_executions

        task1 = self._assert_single_item(tasks, name='task1')

        self._assert_single_item(tasks, name='task2')

        self.assertDictEqual({"my_var": "my branch value"}, task1.published)

    def test_global_publishing_success_access_via_root_context_(self):
        wf_text = """---
        version: '2.0'

        wf:
          tasks:
            task1:
              action: std.echo output="Hi"
              on-success:
                publish:
                  global:
                    my_var: <% task().result %>
                next:
                  - task2

            task2:
              action: std.echo output=<% $.my_var %>
              publish:
                result: <% task().result %>
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf')

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            tasks = wf_ex.task_executions

        self._assert_single_item(tasks, name='task1')
        task2 = self._assert_single_item(tasks, name='task2')

        self.assertDictEqual({'result': 'Hi'}, task2.published)

    def test_global_publishing_error_access_via_root_context(self):
        wf_text = """---
        version: '2.0'

        wf:
          tasks:
            task1:
              action: std.fail
              on-success:
                publish:
                  global:
                    my_var: "We got success"
                next:
                  - task2
              on-error:
                publish:
                  global:
                    my_var: "We got an error"
                next:
                  - task2

            task2:
              action: std.echo output=<% $.my_var %>
              publish:
                result: <% task().result %>
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf')

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            tasks = wf_ex.task_executions

        self._assert_single_item(tasks, name='task1')
        task2 = self._assert_single_item(tasks, name='task2')

        self.assertDictEqual({'result': 'We got an error'}, task2.published)

    def test_global_publishing_success_access_via_function(self):
        wf_text = """---
        version: '2.0'

        wf:
          tasks:
            task1:
              action: std.noop
              on-success:
                publish:
                  branch:
                    my_var: Branch local value
                  global:
                    my_var: Global value
                next:
                  - task2

            task2:
              action: std.noop
              publish:
                local: <% $.my_var %>
                global: <% global(my_var) %>
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf')

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            tasks = wf_ex.task_executions

        self._assert_single_item(tasks, name='task1')
        task2 = self._assert_single_item(tasks, name='task2')

        self.assertDictEqual(
            {
                'local': 'Branch local value',
                'global': 'Global value'
            },
            task2.published
        )

    def test_global_publishing_error_access_via_function(self):
        wf_text = """---
        version: '2.0'

        wf:
          tasks:
            task1:
              action: std.fail
              on-error:
                publish:
                  branch:
                    my_var: Branch local value
                  global:
                    my_var: Global value
                next:
                  - task2

            task2:
              action: std.noop
              publish:
                local: <% $.my_var %>
                global: <% global(my_var) %>
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf')

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            tasks = wf_ex.task_executions

        self._assert_single_item(tasks, name='task1')
        task2 = self._assert_single_item(tasks, name='task2')

        self.assertDictEqual(
            {
                'local': 'Branch local value',
                'global': 'Global value'
            },
            task2.published
        )

    def test_get_published_global(self):
        wf_text = """---
        version: '2.0'

        wf:
          vars:
            var1: 1
            var2: 2

          tasks:
            task1:
              action: std.noop
              on-success:
                publish:
                  global:
                    global_var1: Global value 1
                    global_var2: Global value 2
        """

        wf_service.create_workflows(wf_text)

        wf_ex = self.engine.start_workflow('wf')

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            published_global = (
                data_flow.get_workflow_execution_published_global(wf_ex)
            )

        self.assertDictEqual(
            {
                'global_var1': 'Global value 1',
                'global_var2': 'Global value 2'
            },
            published_global
        )

    def test_linear_data_with_input_expressions(self):
        wf_text = """---
        version: '2.0'

        wf:
          tasks:
            task1:
              action: std.echo
              input:
                output:
                  key1: value1
                  key2: value2
              publish:
                res1: <% task(task1).result %>
              on-success:
                - task2

            task2:
              action: std.echo output=<% $.res1.key2 %>
              publish:
                res2: <% task().result %>
        """

        wf_service.create_workflows(wf_text)

        # Start workflow.
        wf_ex = self.engine.start_workflow('wf')

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            tasks = wf_ex.task_executions

        task1 = self._assert_single_item(
            tasks,
            name='task1',
            state=states.SUCCESS
        )
        task2 = self._assert_single_item(
            tasks,
            name='task2',
            state=states.SUCCESS
        )

        self.assertDictEqual(
            {
                'key1': 'value1',
                'key2': 'value2'
            },
            task1.published['res1']
        )
        self.assertEqual('value2', task2.published['res2'])


class DataFlowTest(test_base.BaseTest):
    def test_get_task_execution_result(self):
        task_ex = models.TaskExecution(
            name='task1',
            spec={
                "version": '2.0',
                'name': 'task1',
                'with-items': 'var in [1]',
                'type': 'direct',
                'action': 'my_action'
            },
            runtime_context={
                'with_items': {'count': 1}
            }
        )

        task_ex.action_executions = [models.ActionExecution(
            name='my_action',
            output={'result': 1},
            accepted=True,
            runtime_context={'index': 0}
        )]

        self.assertEqual([1], data_flow.get_task_execution_result(task_ex))

        task_ex.action_executions.append(models.ActionExecution(
            name='my_action',
            output={'result': 1},
            accepted=True,
            runtime_context={'index': 0}
        ))

        task_ex.action_executions.append(models.ActionExecution(
            name='my_action',
            output={'result': 1},
            accepted=False,
            runtime_context={'index': 0}
        ))

        self.assertEqual(
            [1, 1],
            data_flow.get_task_execution_result(task_ex)
        )

    def test_context_view(self):
        ctx = data_flow.ContextView(
            {
                'k1': 'v1',
                'k11': 'v11',
                'k3': 'v3'
            },
            {
                'k2': 'v2',
                'k21': 'v21',
                'k3': 'v32'
            }
        )

        self.assertIsInstance(ctx, dict)
        self.assertEqual(5, len(ctx))

        self.assertIn('k1', ctx)
        self.assertIn('k11', ctx)
        self.assertIn('k3', ctx)
        self.assertIn('k2', ctx)
        self.assertIn('k21', ctx)

        self.assertEqual('v1', ctx['k1'])
        self.assertEqual('v1', ctx.get('k1'))
        self.assertEqual('v11', ctx['k11'])
        self.assertEqual('v11', ctx.get('k11'))
        self.assertEqual('v3', ctx['k3'])
        self.assertEqual('v2', ctx['k2'])
        self.assertEqual('v2', ctx.get('k2'))
        self.assertEqual('v21', ctx['k21'])
        self.assertEqual('v21', ctx.get('k21'))

        self.assertIsNone(ctx.get('Not existing key'))

        self.assertRaises(exc.MistralError, ctx.update)
        self.assertRaises(exc.MistralError, ctx.clear)
        self.assertRaises(exc.MistralError, ctx.pop, 'k1')
        self.assertRaises(exc.MistralError, ctx.popitem)
        self.assertRaises(exc.MistralError, ctx.__setitem__, 'k5', 'v5')
        self.assertRaises(exc.MistralError, ctx.__delitem__, 'k2')

        self.assertEqual('v1', expr.evaluate('<% $.k1 %>', ctx))
        self.assertEqual('v2', expr.evaluate('<% $.k2 %>', ctx))
        self.assertEqual('v3', expr.evaluate('<% $.k3 %>', ctx))

        # Now change the order of dictionaries and make sure to have
        # a different for key 'k3'.
        ctx = data_flow.ContextView(
            {
                'k2': 'v2',
                'k21': 'v21',
                'k3': 'v32'
            },
            {
                'k1': 'v1',
                'k11': 'v11',
                'k3': 'v3'
            }
        )

        self.assertEqual('v32', expr.evaluate('<% $.k3 %>', ctx))

    def test_context_view_eval_root_with_yaql(self):
        ctx = data_flow.ContextView(
            {'k1': 'v1'},
            {'k2': 'v2'}
        )

        res = expr.evaluate('<% $ %>', ctx)

        self.assertIsNotNone(res)
        self.assertIsInstance(res, dict)
        self.assertEqual(2, len(res))

    def test_context_view_eval_keys(self):
        ctx = data_flow.ContextView(
            {'k1': 'v1'},
            {'k2': 'v2'}
        )

        res = expr.evaluate('<% $.keys() %>', ctx)

        self.assertIsNotNone(res)
        self.assertIsInstance(res, set)
        self.assertEqual(2, len(res))
        self.assertIn('k1', res)
        self.assertIn('k2', res)

    def test_context_view_eval_values(self):
        ctx = data_flow.ContextView(
            {'k1': 'v1'},
            {'k2': 'v2'}
        )

        res = expr.evaluate('<% $.values() %>', ctx)

        self.assertIsNotNone(res)
        self.assertIsInstance(res, list)
        self.assertEqual(2, len(res))
        self.assertIn('v1', res)
        self.assertIn('v2', res)

    def test_context_view_repr(self):
        ctx = data_flow.ContextView(
            {'k1': 'v1'},
            {'k2': 'v2'},
            {3: 3}
        )

        str_repr = str(ctx)

        self.assertIsNotNone(str_repr)
        self.assertFalse(str_repr == "{}")
        self.assertEqual("{'k1': 'v1', 'k2': 'v2', 3: 3}", str_repr)

        ctx = data_flow.ContextView()

        self.assertEqual('{}', str(ctx))

    def test_context_view_as_root_json(self):
        ctx = data_flow.ContextView(
            {'k1': 'v1'},
            {'k2': 'v2'},
        )

        json_str = utils.to_json_str(ctx)

        self.assertIsNotNone(json_str)
        self.assertNotEqual('{}', json_str)

        # We can't use regular dict comparison because key order
        # is not defined.
        self.assertIn('"k1": "v1"', json_str)
        self.assertIn('"k2": "v2"', json_str)

    def test_context_view_as_nested_json(self):
        ctx = data_flow.ContextView(
            {'k1': 'v1'},
            {'k2': 'v2'},
        )

        d = {'root': ctx}

        json_str = utils.to_json_str(d)

        self.assertIsNotNone(json_str)
        self.assertNotEqual('{"root": {}}', json_str)

        # We can't use regular dict comparison because key order
        # is not defined.
        self.assertIn('"k1": "v1"', json_str)
        self.assertIn('"k1": "v1"', json_str)
        self.assertIn('"root"', json_str)



# Copyright 2013 - Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import hashlib
from mistral.db.v2 import api as db_api
from mistral import exceptions as exc
from mistral.lang import parser as spec_parser
from mistral import services
from mistral.utils import safe_yaml
from mistral.workflow import states
from mistral_lib import utils
from oslo_log import log as logging
from stevedore import extension


STD_WF_PATH = 'resources/workflows'
LOG = logging.getLogger(__name__)


def register_preinstalled_workflows(run_in_tx=True):

    extensions = extension.ExtensionManager(
        namespace='mistral.preinstalled_workflows',
        invoke_on_load=True
    )

    for ext in extensions:
        for wf_path in ext.obj:
            register_workflow(run_in_tx, wf_path)


def register_workflow(run_in_tx, wf_path):
    LOG.debug("Registering preinstalled workflow %s", wf_path)

    workflow_definition = open(wf_path).read()

    create_workflows(
        workflow_definition,
        scope='public',
        is_system=True,
        run_in_tx=run_in_tx,
        namespace=''
    )


def _clear_system_workflow_db():
    db_api.delete_workflow_definitions(is_system=True)


def sync_db():
    LOG.debug("Syncing db...")

    with db_api.transaction():
        _clear_system_workflow_db()

        register_preinstalled_workflows(run_in_tx=False)


def create_workflows(definition, scope='private', is_system=False,
                     run_in_tx=True, namespace='', validate=True):
    LOG.debug("Creating workflows...")

    wf_list_spec = spec_parser.get_workflow_list_spec_from_yaml(
        definition,
        validate=services.is_validation_enabled(validate)
    )

    db_wfs = []

    if run_in_tx:
        with db_api.transaction():
            _append_all_workflows(
                definition,
                is_system,
                scope,
                namespace,
                wf_list_spec,
                db_wfs
            )
    else:
        _append_all_workflows(
            definition,
            is_system,
            scope,
            namespace,
            wf_list_spec,
            db_wfs
        )

    return db_wfs


def _append_all_workflows(definition, is_system, scope, namespace,
                          wf_list_spec, db_wfs):
    wfs = wf_list_spec.get_workflows()

    wfs_yaml = safe_yaml.load(definition) if len(wfs) != 1 else None

    for wf_spec in wfs:
        if len(wfs) != 1:
            definition = _cut_wf_definition_from_all(
                wfs_yaml,
                wf_spec.get_name()
            )

        db_wfs.append(
            _create_workflow(
                wf_spec,
                definition,
                scope,
                namespace,
                is_system
            )
        )


def update_workflows(definition, scope='private', identifier=None,
                     namespace='', validate=True):
    LOG.debug("Updating workflows...")

    wf_list_spec = spec_parser.get_workflow_list_spec_from_yaml(
        definition,
        validate=services.is_validation_enabled(validate)
    )

    wfs = wf_list_spec.get_workflows()

    if identifier and len(wfs) > 1:
        raise exc.InputException(
            "More than one workflows are not supported for "
            "update with identifier. [identifier: %s]" %
            identifier
        )

    db_wfs = []

    wfs_yaml = safe_yaml.load(definition) if len(wfs) != 1 else None

    with db_api.transaction():
        for wf_spec in wfs:
            if len(wfs) != 1:
                definition = _cut_wf_definition_from_all(
                    wfs_yaml,
                    wf_spec.get_name()
                )

            db_wfs.append(
                _update_workflow(
                    wf_spec,
                    definition,
                    scope,
                    namespace=namespace,
                    identifier=identifier
                )
            )

    return db_wfs


def update_workflow_execution_env(wf_ex, env):
    if not env:
        return wf_ex

    if wf_ex.state not in [states.IDLE, states.PAUSED, states.ERROR]:
        raise exc.NotAllowedException(
            'Updating env to workflow execution is only permitted if '
            'it is in IDLE, PAUSED, or ERROR state.'
        )

    wf_ex.params['env'] = utils.merge_dicts(wf_ex.params['env'], env)

    return wf_ex


def get_workflow_definition_checksum(definition):
    return hashlib.md5(definition.encode('utf-8')).hexdigest()


def _get_workflow_values(wf_spec, definition, scope, namespace=None,
                         is_system=False):
    values = {
        'name': wf_spec.get_name(),
        'tags': wf_spec.get_tags(),
        'definition': definition,
        'checksum': get_workflow_definition_checksum(definition),
        'spec': wf_spec.to_dict(),
        'scope': scope,
        'namespace': namespace,
        'is_system': is_system
    }

    return values


def _create_workflow(wf_spec, definition, scope, namespace, is_system):
    return db_api.create_workflow_definition(
        _get_workflow_values(wf_spec, definition, scope, namespace, is_system)
    )


def _update_workflow(wf_spec, definition, scope, identifier=None,
                     namespace=''):
    values = _get_workflow_values(wf_spec, definition, scope, namespace)

    return db_api.update_workflow_definition(
        identifier if identifier else values['name'],
        values
    )


def _cut_wf_definition_from_all(wfs_yaml, wf_name):
    return safe_yaml.dump({
        'version': wfs_yaml['version'],
        wf_name: wfs_yaml[wf_name]
    }, sort_keys=False)



# Copyright 2016 - IBM Corp.
# Copyright 2016 Catalyst IT Limited
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_log import log as logging
from pecan import rest
import wsmeext.pecan as wsme_pecan

from mistral.api import access_control as acl
from mistral.api.controllers.v2 import resources
from mistral.api.controllers.v2 import types
from mistral import context as auth_ctx
from mistral.db.v2 import api as db_api
from mistral import exceptions as exc
from mistral.services import triggers
from mistral.utils import rest_utils

LOG = logging.getLogger(__name__)

UPDATE_NOT_ALLOWED = ['exchange', 'topic', 'event']
CREATE_MANDATORY = set(['exchange', 'topic', 'event', 'workflow_id'])


class EventTriggersController(rest.RestController):
    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.EventTrigger, types.uuid, types.uniquelist)
    def get(self, id, fields=''):
        """Returns the specified event_trigger."""
        acl.enforce('event_triggers:get', auth_ctx.ctx())

        LOG.debug('Fetch event trigger [id=%s]', id)

        if fields and 'id' not in fields:
            fields.insert(0, 'id')

        # Use retries to prevent possible failures.
        r = rest_utils.create_db_retry_object()
        db_model = r.call(db_api.get_event_trigger, id, fields=fields)
        if fields:
            return resources.EventTrigger.from_tuples(zip(fields, db_model))
        return resources.EventTrigger.from_db_model(db_model, fields=fields)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.EventTrigger, body=resources.EventTrigger,
                         status_code=201)
    def post(self, event_trigger):
        """Creates a new event trigger."""
        acl.enforce('event_triggers:create', auth_ctx.ctx())

        values = event_trigger.to_dict()
        input_keys = [k for k in values if values[k]]

        if CREATE_MANDATORY - set(input_keys):
            raise exc.EventTriggerException(
                "Params %s must be provided for creating event trigger." %
                CREATE_MANDATORY
            )

        if values.get('scope') == 'public':
            acl.enforce('event_triggers:create:public', auth_ctx.ctx())

        LOG.debug('Create event trigger: %s', values)

        db_model = rest_utils.rest_retry_on_db_error(
            triggers.create_event_trigger
        )(
            name=values.get('name', ''),
            exchange=values.get('exchange'),
            topic=values.get('topic'),
            event=values.get('event'),
            workflow_id=values.get('workflow_id'),
            scope=values.get('scope'),
            workflow_input=values.get('workflow_input'),
            workflow_params=values.get('workflow_params'),
        )

        return resources.EventTrigger.from_db_model(db_model)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.EventTrigger, types.uuid,
                         body=resources.EventTrigger)
    def put(self, id, event_trigger):
        """Updates an existing event trigger.

        The exchange, topic and event can not be updated. The right way to
        change them is to delete the event trigger first, then create a new
        event trigger with new params.
        """
        acl.enforce('event_triggers:update', auth_ctx.ctx())

        values = event_trigger.to_dict()

        for field in UPDATE_NOT_ALLOWED:
            if values.get(field):
                raise exc.EventTriggerException(
                    "Can not update fields %s of event trigger." %
                    UPDATE_NOT_ALLOWED
                )

        LOG.debug('Update event trigger: [id=%s, values=%s]', id, values)

        @rest_utils.rest_retry_on_db_error
        def _update_event_trigger():
            with db_api.transaction():
                # ensure that event trigger exists
                db_api.get_event_trigger(id)

                return triggers.update_event_trigger(id, values)

        db_model = _update_event_trigger()

        return resources.EventTrigger.from_db_model(db_model)

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(None, types.uuid, status_code=204)
    def delete(self, id):
        """Delete event trigger."""
        acl.enforce('event_triggers:delete', auth_ctx.ctx())

        LOG.debug("Delete event trigger [id=%s]", id)

        @rest_utils.rest_retry_on_db_error
        def _delete_event_trigger():
            with db_api.transaction():
                event_trigger = db_api.get_event_trigger(id)

                triggers.delete_event_trigger(event_trigger.to_dict())

        _delete_event_trigger()

    @rest_utils.wrap_wsme_controller_exception
    @wsme_pecan.wsexpose(resources.EventTriggers, types.uuid, int,
                         types.uniquelist, types.list, types.uniquelist,
                         bool, types.jsontype)
    def get_all(self, marker=None, limit=None, sort_keys='created_at',
                sort_dirs='asc', fields='', all_projects=False, **filters):
        """Return all event triggers."""
        acl.enforce('event_triggers:list', auth_ctx.ctx())

        if all_projects:
            acl.enforce('event_triggers:list:all_projects', auth_ctx.ctx())

        LOG.debug(
            "Fetch event triggers. marker=%s, limit=%s, sort_keys=%s, "
            "sort_dirs=%s, fields=%s, all_projects=%s, filters=%s", marker,
            limit, sort_keys, sort_dirs, fields, all_projects, filters
        )

        return rest_utils.get_all(
            resources.EventTriggers,
            resources.EventTrigger,
            db_api.get_event_triggers,
            db_api.get_event_trigger,
            resource_function=None,
            marker=marker,
            limit=limit,
            sort_keys=sort_keys,
            sort_dirs=sort_dirs,
            fields=fields,
            all_projects=all_projects,
            **filters
        )



# Copyright 2015 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Kilo

Revision ID: 002
Revises: 001
Create Date: 2015-04-30 16:15:34.737030

"""

# revision identifiers, used by Alembic.
revision = '002'
down_revision = '001'

from alembic import op
import sqlalchemy as sa

from mistral.db.sqlalchemy import types as st


def upgrade():
    op.drop_table('tasks')
    op.drop_table('workflow_executions')
    op.drop_table('workbooks')
    op.drop_table('triggers')
    op.add_column(
        'cron_triggers_v2',
        sa.Column('workflow_params', st.JsonEncoded(), nullable=True)
    )
    op.add_column(
        'cron_triggers_v2',
        sa.Column('workflow_params_hash', sa.CHAR(length=64), nullable=True)
    )
    op.create_unique_constraint(
        None,
        'cron_triggers_v2',
        ['workflow_input_hash', 'workflow_name', 'pattern',
         'project_id', 'workflow_params_hash']
    )



# Copyright 2013 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
# Copyright 2016 - Brocade Communications Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import warnings

from oslo_log import log as logging
from oslo_serialization import jsonutils
import yaml

from mistral.db import utils as db_utils
from mistral.db.v2 import api as db_api
from mistral.utils import filter_utils
from mistral_lib import utils as ml_utils

# Additional YAQL/Jinja functions provided by Mistral out of the box.
# If a function name ends with underscore then it doesn't need to pass
# the name of the function when context registers it.

LOG = logging.getLogger(__name__)


def env_(context):
    return context['__env']


@db_utils.tx_cached(ignore_args='context')
def executions_(context, id=None, root_execution_id=None, state=None,
                from_time=None, to_time=None):
    filter_ = {}

    if id is not None:
        filter_ = filter_utils.create_or_update_filter('id', id, "eq", filter_)

    if root_execution_id is not None:
        filter_ = filter_utils.create_or_update_filter(
            'root_execution_id',
            root_execution_id,
            'eq',
            filter_
        )

    if state is not None:
        filter_ = filter_utils.create_or_update_filter(
            'state',
            state,
            'eq',
            filter_
        )

    if from_time is not None:
        filter_ = filter_utils.create_or_update_filter(
            'created_at',
            from_time,
            'gte',
            filter_
        )

    if to_time is not None:
        filter_ = filter_utils.create_or_update_filter(
            'created_at',
            to_time,
            'lt',
            filter_
        )

    return db_api.get_workflow_executions(**filter_)


@db_utils.tx_cached(ignore_args='context')
def execution_(context):
    wf_ex = db_api.get_workflow_execution(context['__execution']['id'])

    return {
        'id': wf_ex.id,
        'name': wf_ex.name,
        'spec': wf_ex.spec,
        'input': wf_ex.input,
        'params': wf_ex.params,
        'created_at': wf_ex.created_at.isoformat(' '),
        'updated_at': wf_ex.updated_at.isoformat(' '),
        'root_execution_id': wf_ex.root_execution_id
    }


def json_pp_(context, data=None):
    warnings.warn(
        "json_pp was deprecated in Queens and will be removed in the S cycle. "
        "The json_dump expression function can be used for outputting JSON",
        DeprecationWarning
    )

    return jsonutils.dumps(
        data or context,
        indent=4
    ).replace("\\n", "\n").replace(" \n", "\n")


def json_dump_(context, data):
    return jsonutils.dumps(data, indent=4)


def yaml_dump_(context, data):
    return yaml.safe_dump(data, default_flow_style=False)


@db_utils.tx_cached(ignore_args='context')
def task_(context, task_name=None):
    # This section may not exist in a context if it's calculated not in
    # task scope.
    cur_task = context['__task_execution']

    # 1. If task_name is empty it's 'task()' use case, we need to get the
    # current task.
    # 2. if task_name is not empty but it's equal to the current task name
    # we need to take exactly the current instance of this task. Otherwise
    # there may be ambiguity if there are many tasks with this name.
    # 3. In other case we just find a task in DB by the given name.
    if cur_task and (not task_name or cur_task['name'] == task_name):
        task_ex = db_api.get_task_execution(cur_task['id'])
    else:
        task_execs = db_api.get_task_executions(
            workflow_execution_id=context['__execution']['id'],
            name=task_name
        )

        # TODO(rakhmerov): Account for multiple executions (i.e. in case of
        # cycles).
        task_ex = task_execs[-1] if len(task_execs) > 0 else None

    if not task_ex:
        LOG.warning(
            "Task '%s' not found by the task() expression function",
            task_name
        )
        return None

    # We don't use to_dict() db model method because not all fields
    # make sense for user.
    return _convert_to_user_model(task_ex)


def _should_pass_filter(t, state, flat):
    # Start from assuming all is true, check only if needed.
    state_match = True
    flat_match = True

    if state:
        state_match = t['state'] == state

    # To break cyclic dependency.
    from mistral.lang.v2 import tasks as lang_tasks

    if flat:
        is_action = t['type'] == lang_tasks.ACTION_TASK_TYPE

        if not is_action:
            nested_execs = db_api.get_workflow_executions(
                task_execution_id=t.id
            )

            for n in nested_execs:
                flat_match = flat_match and n.state != t.state

    return state_match and flat_match


def _get_tasks_from_db(workflow_execution_id=None, recursive=False, state=None,
                       flat=False):
    task_execs = []
    nested_task_exs = []

    kwargs = {}

    if workflow_execution_id:
        kwargs['workflow_execution_id'] = workflow_execution_id

    # We can't add state to query if we want to filter by workflow_execution_id
    # recursively. There might be a workflow_execution in one state with a
    # nested workflow execution that has a task in the desired state until we
    # have an optimization for queering all workflow executions under a given
    # top level workflow execution, this is the way to go.
    if state and not (workflow_execution_id and recursive):
        kwargs['state'] = state

    task_execs.extend(db_api.get_task_executions(**kwargs))

    # To break cyclic dependency.
    from mistral.lang.v2 import tasks as lang_tasks

    # If it is not recursive no need to check nested workflows.
    # If there is no workflow execution id, we already have all we need, and
    # doing more queries will just create duplication in the results.
    if recursive and workflow_execution_id:
        for t in task_execs:
            if t.type == lang_tasks.WORKFLOW_TASK_TYPE:
                # Get nested workflow execution that matches the task.
                nested_workflow_executions = db_api.get_workflow_executions(
                    task_execution_id=t.id
                )

                # There might be zero nested executions.
                for nested_workflow_execution in nested_workflow_executions:
                    nested_task_exs.extend(
                        _get_tasks_from_db(
                            nested_workflow_execution.id,
                            recursive,
                            state,
                            flat
                        )
                    )

    if state or flat:
        # Filter by state and flat.
        task_execs = [
            t for t in task_execs if _should_pass_filter(t, state, flat)
        ]

    # The nested tasks were already filtered, since this is a recursion.
    task_execs.extend(nested_task_exs)

    return task_execs


@db_utils.tx_cached(ignore_args='context')
def tasks_(context, workflow_execution_id=None, recursive=False, state=None,
           flat=False):
    task_execs = _get_tasks_from_db(
        workflow_execution_id,
        recursive,
        state,
        flat
    )

    # Convert task_execs to user model and return.
    return [_convert_to_user_model(t) for t in task_execs]


def _convert_to_user_model(task_ex):
    # Importing data_flow in order to break cycle dependency between modules.
    from mistral.workflow import data_flow

    # We don't use to_dict() db model method because not all fields
    # make sense for user.
    return {
        'id': task_ex.id,
        'name': task_ex.name,
        'spec': task_ex.spec,
        'state': task_ex.state,
        'state_info': task_ex.state_info,
        'result': data_flow.get_task_execution_result(task_ex),
        'published': task_ex.published,
        'type': task_ex.type,
        'workflow_execution_id': task_ex.workflow_execution_id,
        'created_at': task_ex.created_at.isoformat(' '),
        'updated_at': task_ex.updated_at.isoformat(' ')
        if task_ex.updated_at is not None else None
    }


def uuid_(context=None):
    return ml_utils.generate_unicode_uuid()


def global_(context, var_name):
    wf_ex = db_api.get_workflow_execution(context['__execution']['id'])

    return wf_ex.context.get(var_name)


def json_parse_(context, data):
    return jsonutils.loads(data)


def yaml_parse_(context, data):
    return yaml.safe_load(data)



# Copyright 2015 - Mirantis, Inc.
# Copyright 2016 - Brocade Communications Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import eventlet
import functools
from oslo_config import cfg
from oslo_log import log as logging
from osprofiler import profiler

from mistral import context
from mistral.db import utils as db_utils
from mistral.db.v2 import api as db_api
from mistral_lib import utils


"""
This module contains a mini framework for scheduling operations while
performing transactional processing of a workflow event such as
completing a workflow action. The scheduled operations will run after
the main DB transaction, in a new transaction, if needed.
"""

LOG = logging.getLogger(__name__)


_THREAD_LOCAL_NAME = "__operation_queue_thread_local"


def _prepare():
    # Register two queues: transactional and non transactional operations.
    utils.set_thread_local(_THREAD_LOCAL_NAME, (list(), list()))


def _clear():
    utils.set_thread_local(_THREAD_LOCAL_NAME, None)


def register_operation(func, args=None, in_tx=False):
    """Register an operation."""

    _get_queues()[0 if in_tx else 1].append((func, args or []))


def _get_queues():
    queues = utils.get_thread_local(_THREAD_LOCAL_NAME)

    if queues is None:
        raise RuntimeError(
            'Operation queue is not initialized for the current thread.'
            ' Most likely some engine method is not decorated with'
            ' operation_queue.run()'
        )

    return queues


def run(func):
    """Decorator that runs all operations registered in the operation queue.

    Various engine methods may register such operations. All such methods must
    be decorated with this decorator.
    """
    @functools.wraps(func)
    def decorate(*args, **kw):
        _prepare()

        try:
            res = func(*args, **kw)

            queues = _get_queues()

            tx_queue = queues[0]
            non_tx_queue = queues[1]

            if not tx_queue and not non_tx_queue:
                return res

            auth_ctx = context.ctx() if context.has_ctx() else None

            def _within_new_thread():
                # This is a new thread so we need to init a profiler again.
                if cfg.CONF.profiler.enabled:
                    profiler.init(cfg.CONF.profiler.hmac_keys)

                old_auth_ctx = context.ctx() if context.has_ctx() else None

                context.set_ctx(auth_ctx)

                try:
                    if tx_queue:
                        _process_tx_queue(tx_queue)

                    if non_tx_queue:
                        _process_non_tx_queue(non_tx_queue)
                finally:
                    context.set_ctx(old_auth_ctx)

            eventlet.spawn(_within_new_thread)
        finally:
            _clear()

        return res

    return decorate


@db_utils.retry_on_db_error
@run
def _process_tx_queue(queue):
    with db_api.transaction():
        for func, args in queue:
            try:
                func(*args)
            except Exception:
                LOG.exception("Failed to run transactional engine operation.")

                raise


def _process_non_tx_queue(queue):
    for func, args in queue:
        try:
            func(*args)
        except Exception:
            LOG.exception("Failed to run non-transactional engine operation.")



#  Copyright 2019 - Nokia Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

from mistral.services import workflows as wf_service
from mistral.tests.unit.api import base
from mistral.tests.unit.engine import base as engine_base

WF_TEXT = """---
version: '2.0'

wf:
  tasks:
       task1:
         action: std.noop
         on-success:
           publish:
             branch:
               my_var: Branch local value
             global:
               my_var: Global value
           next:
             - task2

       task2:
         action: std.noop
         publish:
           local: <% $.my_var %>
           global: <% global(my_var) %>

"""


def _find_task(task_name, tasks):
    return next(
        (
            task
            for task in tasks
            if task['name'] == task_name
        ), None
    )


class TestGlobalPublish(base.APITest, engine_base.EngineTestCase):
    def setUp(self):
        super(TestGlobalPublish, self).setUp()

        wf_service.create_workflows(WF_TEXT)

        wf_ex = self.engine.start_workflow('wf')

        self.await_workflow_success(wf_ex.id)

        self.wf_id = wf_ex.id

    def test_global_publish_in_task_exec(self):
        resp = self.app.get('/v2/tasks/')

        tasks = resp.json['tasks']
        task = _find_task('task1', tasks)

        self.assertIsNotNone(task, 'task1 not found')

        resp = self.app.get('/v2/tasks/%s/' % task['id'])

        self.assert_for_published_global(resp)

    def test_global_publish_in_wf_exec(self):
        resp = self.app.get('/v2/executions/%s/' % self.wf_id)

        self.assert_for_published_global(resp)

    def assert_for_published_global(self, resp):
        self.assertEqual(200, resp.status_int)
        self.assertEqual(
            resp.json['published_global'],
            '{"my_var": "Global value"}'
        )



# Copyright 2015 Huawei Technologies Co., Ltd.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
import json

from mistral.utils import filter_utils
from oslo_utils import uuidutils
from wsme import types as wtypes

from mistral import exceptions as exc


class ListType(wtypes.UserType):
    """A simple list type."""

    basetype = wtypes.text
    name = 'list'

    @staticmethod
    def validate(value):
        """Validate and convert the input to a ListType.

        :param value: A comma separated string of values
        :returns: A list of values.
        """
        items = [v.strip().lower() for v in str(value).split(',')]

        # remove empty items.
        return [x for x in items if x]

    @staticmethod
    def frombasetype(value):
        return ListType.validate(value) if value is not None else None


class UniqueListType(ListType):
    """A simple list type with no duplicate items."""

    name = 'uniquelist'

    @staticmethod
    def validate(value):
        """Validate and convert the input to a UniqueListType.

        :param value: A comma separated string of values.
        :returns: A list with no duplicate items.
        """
        items = ListType.validate(value)

        seen = set()

        return [x for x in items if not (x in seen or seen.add(x))]

    @staticmethod
    def frombasetype(value):
        return UniqueListType.validate(value) if value is not None else None


class UuidType(wtypes.UserType):
    """A simple UUID type.

    The builtin UuidType class in wsme.types doesn't work properly with pecan.
    """

    basetype = wtypes.text
    name = 'uuid'

    @staticmethod
    def validate(value):
        _, data = filter_utils.extract_filter_type_and_value(value)
        if not uuidutils.is_uuid_like(data):
            raise exc.InputException(
                "Expected a uuid but received %s." % data
            )

        return data

    @staticmethod
    def frombasetype(value):
        return UuidType.validate(value) if value is not None else None


class JsonType(wtypes.UserType):
    """A simple JSON type."""

    basetype = wtypes.text
    name = 'json'

    def validate(self, value):
        if not value:
            return {}

        if not isinstance(value, dict):
            raise exc.InputException(
                'JsonType field value must be a dictionary [actual=%s]' % value
            )

        return value

    def frombasetype(self, value):
        if isinstance(value, dict):
            return value
        try:
            return json.loads(value) if value is not None else None
        except TypeError as e:
            raise ValueError(e)

    def tobasetype(self, value):
        # Value must be a dict.
        return json.dumps(value) if value is not None else None


uuid = UuidType()
list = ListType()
uniquelist = UniqueListType()
jsontype = JsonType()



# Copyright 2014 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from mistral.lang import types
from mistral.lang.v2 import base
from mistral_lib import utils


class ActionSpec(base.BaseSpec):
    # See http://json-schema.org
    _schema = {
        "type": "object",
        "properties": {
            "base": types.NONEMPTY_STRING,
            "base-input": types.NONEMPTY_DICT,
            "input": types.UNIQUE_STRING_OR_ONE_KEY_DICT_LIST,
            "output": types.ANY_NULLABLE,
        },
        "required": ["base"],
        "additionalProperties": False
    }

    def __init__(self, data, validate):
        super(ActionSpec, self).__init__(data, validate)

        self._name = data['name']
        self._description = data.get('description')
        self._tags = data.get('tags', [])
        self._base = data['base']
        self._base_input = data.get('base-input', {})
        self._input = utils.get_dict_from_entries(data.get('input', []))
        self._output = data.get('output')

        self._base, _input = self._parse_cmd_and_input(self._base)

        utils.merge_dicts(self._base_input, _input)

    def validate_schema(self):
        super(ActionSpec, self).validate_schema()

        # Validate YAQL expressions.
        inline_params = self._parse_cmd_and_input(self._data.get('base'))[1]
        self.validate_expr(inline_params)

        self.validate_expr(self._data.get('base-input', {}))

        if isinstance(self._data.get('output'), str):
            self.validate_expr(self._data.get('output'))

    def get_name(self):
        return self._name

    def get_description(self):
        return self._description

    def get_tags(self):
        return self._tags

    def get_base(self):
        return self._base

    def get_base_input(self):
        return self._base_input

    def get_input(self):
        return self._input

    def get_output(self):
        return self._output


class ActionSpecList(base.BaseSpecList):
    item_class = ActionSpec


class ActionListSpec(base.BaseListSpec):
    item_class = ActionSpec

    def get_actions(self):
        return self.get_items()



#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
from alembic import context
from logging import config as c
from oslo_utils import importutils
from sqlalchemy import create_engine
from sqlalchemy import pool

from mistral.db.sqlalchemy import model_base


importutils.try_import('mistral.db.v2.sqlalchemy.models')

# This is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config
mistral_config = config.mistral_config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
c.fileConfig(config.config_file_name)

# Add your model's MetaData object here for 'autogenerate' support.
target_metadata = model_base.MistralSecureModelBase.metadata


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    context.configure(url=mistral_config.database.connection)

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    engine = create_engine(
        mistral_config.database.connection,
        poolclass=pool.NullPool
    )

    connection = engine.connect()
    context.configure(
        connection=connection,
        target_metadata=target_metadata
    )

    try:
        with context.begin_transaction():
            context.run_migrations()
    finally:
        connection.close()


if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



# Copyright 2020 Nokia Software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""create new tables for the dynamic actions and code sources

Revision ID: 001
Revises: None
Create Date: 2020-09-30 12:02:51.935368

"""

# revision identifiers, used by Alembic.
revision = '040'
down_revision = '039'

from alembic import op
from mistral.db.sqlalchemy import types as st
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'code_sources',

        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('project_id', sa.String(length=80), nullable=True),
        sa.Column('namespace', sa.String(length=255), nullable=True),
        sa.Column('content', sa.TEXT, nullable=False),
        sa.Column('version', sa.Integer, nullable=False),
        sa.Column('tags', st.JsonEncoded(), nullable=True),
        sa.Column('scope', sa.String(length=80), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),

        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('name', 'namespace', 'project_id'),

        sa.Index('code_sources_project_id', 'project_id'),
        sa.Index('code_sources_scope', 'scope')
    )

    op.create_table(
        'dynamic_action_definitions',

        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=255), nullable=False),
        sa.Column('class_name', sa.String(length=255), nullable=False),
        sa.Column('scope', sa.String(length=80), nullable=True),
        sa.Column('project_id', sa.String(length=80), nullable=True),
        sa.Column('code_source_id', sa.String(length=36), nullable=False),
        sa.Column('code_source_name', sa.String(length=255), nullable=False),
        sa.Column('namespace', sa.String(length=255), nullable=True),
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),

        sa.PrimaryKeyConstraint('id'),
        sa.ForeignKeyConstraint(
            ['code_source_id'],
            ['code_sources.id'],
            ondelete='CASCADE'
        ),

        sa.UniqueConstraint('name', 'namespace', 'project_id'),

        sa.Index('dynamic_action_definitions_project_id', 'project_id'),
        sa.Index('dynamic_action_definitions_scope', 'scope'),

    )



# Copyright 2018 Nokia Networks. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslotest import base

from mistral.api.controllers.v2 import resources


class TestResourceList(base.BaseTestCase):

    def test_next_link_correctness(self):
        task = resources.Task.sample()

        result = resources.Tasks.convert_with_links(
            resources=[task],
            limit=1,
            url='https://localhost:8080',
            sort_keys='created_at,id',
            sort_dirs='asc,asc',
            fields='',
            state='eq:RUNNING'
        )

        next_link = result.next

        self.assertIn('state=eq:RUNNING', next_link)
        self.assertIn('sort_keys=created_at,id', next_link)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from mistral.rpc.kombu import kombu_hosts
from mistral.tests.unit import base
from oslo_config import cfg

import functools
import oslo_messaging


HOST_1 = 'rabbitmq_1'
PORT_1 = 5671
HOST_2 = 'rabbitmq_2'
PORT_2 = 5672
USER_1 = 'u_mistral_1'
PASSWORD_1 = 'p_mistral_1'
USER_2 = 'u_mistral_2'
PASSWORD_2 = 'p_mistral_2'
VIRTUAL_HOST_1 = 'vhost_1'
VIRTUAL_HOST_2 = 'vhost_2'


class KombuHostsTest(base.BaseTest):

    def setUp(self):
        super(KombuHostsTest, self).setUp()

        # Oslo messaging set a default config option
        oslo_messaging.get_transport(cfg.CONF)

    def assert_transports_host(self, expected, result):
        sorted_by_host = functools.partial(sorted, key=lambda x: x.hostname)

        self.assertListEqual(sorted_by_host(expected), sorted_by_host(result))

    def test_transport_url(self):
        self.override_config(
            'transport_url',
            'rabbit://{user}:{password}@{host}:{port}/{virtual_host}'.format(
                user=USER_1, port=PORT_1, host=HOST_1,
                password=PASSWORD_1,
                virtual_host=VIRTUAL_HOST_1
            ))

        hosts = kombu_hosts.KombuHosts(cfg.CONF)

        self.assertEqual(VIRTUAL_HOST_1, hosts.virtual_host)
        self.assert_transports_host([oslo_messaging.TransportHost(
            hostname=HOST_1,
            port=PORT_1,
            username=USER_1,
            password=PASSWORD_1,
        )], hosts.hosts)

    def test_transport_url_multiple_hosts(self):
        self.override_config(
            'transport_url',
            'rabbit://{user_1}:{password_1}@{host_1}:{port_1},'
            '{user_2}:{password_2}@{host_2}:{port_2}/{virtual_host}'.format(
                user_1=USER_1,
                password_1=PASSWORD_1,
                port_1=PORT_1,
                host_1=HOST_1,
                user_2=USER_2,
                password_2=PASSWORD_2,
                host_2=HOST_2,
                port_2=PORT_2,
                virtual_host=VIRTUAL_HOST_1
            ))

        hosts = kombu_hosts.KombuHosts(cfg.CONF)

        self.assertEqual(VIRTUAL_HOST_1, hosts.virtual_host)
        self.assert_transports_host(
            [
                oslo_messaging.TransportHost(
                    hostname=HOST_1,
                    port=PORT_1,
                    username=USER_1,
                    password=PASSWORD_1
                ),
                oslo_messaging.TransportHost(
                    hostname=HOST_2,
                    port=PORT_2,
                    username=USER_2,
                    password=PASSWORD_2
                )
            ],
            hosts.hosts
        )



# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import copy


from mistral import exceptions as exc
from mistral.lang import parser as spec_parser
from mistral.tests.unit import base
from mistral.utils import safe_yaml
from mistral_lib import utils


class WorkflowSpecValidationTestCase(base.BaseTest):

    def __init__(self, *args, **kwargs):
        super(WorkflowSpecValidationTestCase, self).__init__(*args, **kwargs)

        # The relative resource path is ./mistral/tests/resources/workbook/v2.
        self._resource_path = 'workbook/v2'

        self._spec_parser = spec_parser.get_workflow_list_spec_from_yaml

        self._dsl_blank = {
            'version': '2.0',
            'test': {
                'type': 'direct'
            }
        }

        self._dsl_tasks = {
            'get': {
                'action': 'std.http',
                'input': {
                    'url': 'https://www.openstack.org'
                }
            },
            'echo': {
                'action': 'std.echo',
                'input': {
                    'output': 'This is a test.'
                }
            },
            'email': {
                'action': 'std.email',
                'input': {
                    'from_addr': 'mistral@example.com',
                    'to_addrs': ['admin@example.com'],
                    'subject': 'Test',
                    'body': 'This is a test.',
                    'smtp_server': 'localhost',
                    'smtp_password': 'password'
                }
            }
        }

    def _parse_dsl_spec(self, dsl_file=None, add_tasks=False,
                        changes=None, expect_error=False):
        if dsl_file and add_tasks:
            raise Exception('The add_tasks option is not a valid '
                            'combination with the dsl_file option.')

        if dsl_file:
            dsl_yaml = base.get_resource(self._resource_path + '/' + dsl_file)

            if changes:
                dsl_dict = safe_yaml.safe_load(dsl_yaml)
                utils.merge_dicts(dsl_dict, changes)
                dsl_yaml = safe_yaml.safe_dump(dsl_dict,
                                               default_flow_style=False)
        else:
            dsl_dict = copy.deepcopy(self._dsl_blank)

            if add_tasks:
                dsl_dict['test']['tasks'] = copy.deepcopy(self._dsl_tasks)

            if changes:
                utils.merge_dicts(dsl_dict, changes)

            dsl_yaml = safe_yaml.safe_dump(dsl_dict, default_flow_style=False)

        if not expect_error:
            return self._spec_parser(dsl_yaml)
        else:
            return self.assertRaises(
                exc.DSLParsingException,
                self._spec_parser,
                dsl_yaml
            )


class WorkbookSpecValidationTestCase(WorkflowSpecValidationTestCase):

    def __init__(self, *args, **kwargs):
        super(WorkbookSpecValidationTestCase, self).__init__(*args, **kwargs)

        self._spec_parser = spec_parser.get_workbook_spec_from_yaml

        self._dsl_blank = {
            'version': '2.0',
            'name': 'test_wb'
        }

    def _parse_dsl_spec(self, dsl_file=None, add_tasks=False,
                        changes=None, expect_error=False):
        return super(WorkbookSpecValidationTestCase, self)._parse_dsl_spec(
            dsl_file=dsl_file, add_tasks=False, changes=changes,
            expect_error=expect_error)



# Copyright 2020 Nokia Software.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Update type of 'description' field for
   table 'workflow_executions_v2' to text

Revision ID: 041
Revises: 040
Create Date: 2022-04-19

"""

# revision identifiers, used by Alembic.
revision = '041'
down_revision = '040'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.alter_column(
        'workflow_executions_v2',
        column_name='description', type_=sa.Text(),
        existing_type=sa.String(length=255)
    )



#  Copyright 2019 - Nokia Corporation
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.


from time import time
from uuid import UUID
from uuid import uuid5


def my_function_(context):
    """Generate a UUID using the execution ID and the clock."""
    # fetch the current workflow execution ID found in the context
    execution_id = context['__execution']['id']

    time_str = str(time())
    execution_uuid = UUID(execution_id)
    return uuid5(execution_uuid, time_str)



# Copyright 2020 Nokia Software.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#

#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from oslo_policy import policy

from mistral.policies import base

ACTIONS = 'dynamic_actions:%s'
BASE_PATH = '/v2/dynamic_actions'

rules = [
    policy.DocumentedRuleDefault(
        name=ACTIONS % 'create',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description='Create a new dynamic action.',
        operations=[
            {
                'path': BASE_PATH,
                'method': 'POST'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=ACTIONS % 'delete',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description='Delete the named dynamic action.',
        operations=[
            {
                'path': BASE_PATH,
                'method': 'DELETE'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=ACTIONS % 'get',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description='Return the named dynamic action.',
        operations=[
            {
                'path': BASE_PATH + '/{action_id}',
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=ACTIONS % 'list',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description='Return all dynamic actions.',
        operations=[
            {
                'path': BASE_PATH,
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name=ACTIONS % 'update',
        check_str=base.RULE_ADMIN_OR_OWNER,
        description='Update one or more dynamic actions.',
        operations=[
            {
                'path': BASE_PATH,
                'method': 'PUT'
            }
        ]
    )
]


def list_rules():
    return rules



#!/usr/bin/env python

import sys


try:
    wf_name = str(sys.argv[1])
    branch_cnt = int(sys.argv[2])
    branch_depth = int(sys.argv[3])

    add_join = len(sys.argv) > 4
except:
    raise ValueError(
        'Usage: <script_name> workflow_name'
        ' number_of_parallel_branches branch_depth add_join'
    )


f = open('%s.mist' % wf_name, 'w')

# Writing a workflow header to the file.

f.write('---\n')
f.write("version: '2.0'\n\n")

f.write("%s:\n" % wf_name)
f.write("  tasks:\n")

# 1. First starting task.

f.write("    task_1:\n")
f.write("      action: std.noop\n")
f.write("      on-success:\n")

for branch_num in range(1, branch_cnt + 1):
    f.write("        - task_%s_1\n" % branch_num)

# 2. Branch tasks.

for branch_num in range(1, branch_cnt + 1):
    for task_num in range(1, branch_depth + 1):
        f.write("    task_%s_%s:\n" % (branch_num, task_num))
        f.write("      action: std.noop\n")

        if task_num < branch_depth:
            f.write("      on-success: task_%s_%s\n" % (branch_num, task_num + 1))
        elif add_join:
            f.write("      on-success: task_join\n")

# 3. The last "join" task, if needed.

if add_join:
    f.write("    task_join:\n")
    f.write("      join: all")

f.close()

print("Workflow '%s' is created." % wf_name)



# Copyright 2016 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""add event triggers table

Revision ID: 012
Revises: 011
Create Date: 2016-03-04 09:49:52.481791

"""

# revision identifiers, used by Alembic.
revision = '012'
down_revision = '011'

from alembic import op
import sqlalchemy as sa

from mistral.db.sqlalchemy import types as st


def upgrade():
    op.create_table(
        'event_triggers_v2',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('project_id', sa.String(length=80), nullable=True),
        sa.Column('scope', sa.String(length=80), nullable=True),
        sa.Column('id', sa.String(length=36), nullable=False),
        sa.Column('name', sa.String(length=200), nullable=True),
        sa.Column('workflow_id', sa.String(length=36), nullable=False),
        sa.Column('exchange', sa.String(length=80), nullable=False),
        sa.Column('topic', sa.String(length=80), nullable=False),
        sa.Column('event', sa.String(length=80), nullable=False),
        sa.Column('workflow_params', st.JsonEncoded(), nullable=True),
        sa.Column('workflow_input', st.JsonEncoded(), nullable=True),
        sa.Column('trust_id', sa.String(length=80), nullable=True),

        sa.ForeignKeyConstraint(
            ['workflow_id'],
            ['workflow_definitions_v2.id'],
        ),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint(
            'exchange',
            'topic',
            'event',
            'workflow_id',
            'project_id'
        ),
        sa.Index(
            'event_triggers_v2_project_id_workflow_id',
            'project_id', 'workflow_id'
        )
    )



# Copyright 2019 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Namespace support postgresql

Revision ID: 035
Revises: 034
Create Date: 2019-08-01 15:48:34.115639

"""

# revision identifiers, used by Alembic.
revision = '035'
down_revision = '034'

from alembic import op
from sqlalchemy.engine import reflection


def upgrade():

    inspect = reflection.Inspector.from_engine(op.get_bind())

    unique_constraints = [
        unique_constraint['name'] for unique_constraint in
        inspect.get_unique_constraints('workflow_definitions_v2')
    ]

    if 'workflow_definitions_v2_name_project_id_key' in unique_constraints:
        op.drop_constraint('workflow_definitions_v2_name_project_id_key',
                           table_name='workflow_definitions_v2')



# Copyright 2013 - Mirantis, Inc.
# Copyright 2016 - Brocade Communications Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import base64

from mistral_lib.actions import context as lib_ctx
from mistral_lib import serialization
from oslo_config import cfg
from oslo_context import context as oslo_context
from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_serialization import jsonutils
from osprofiler import profiler
import pecan
from pecan import hooks

from mistral import auth
from mistral import exceptions as exc
from mistral_lib import utils


LOG = logging.getLogger(__name__)

CONF = cfg.CONF

_CTX_THREAD_LOCAL_NAME = "MISTRAL_APP_CTX_THREAD_LOCAL"
ALLOWED_WITHOUT_AUTH = [
    '/',
    '/info',
    '/v2/',
    '/workflowv2/',
    '/workflowv2/v2/'
]


class MistralContext(oslo_context.RequestContext):

    FROM_DICT_EXTRA_KEYS = ['root_execution_id']

    def __init__(self, auth_uri=None, auth_cacert=None, insecure=False,
                 service_catalog=None, region_name=None, is_trust_scoped=False,
                 redelivered=False, expires_at=None, trust_id=None,
                 is_target=False, root_execution_id=None, **kwargs):
        self.auth_uri = auth_uri
        self.auth_cacert = auth_cacert
        self.insecure = insecure
        self.service_catalog = service_catalog
        self.region_name = region_name
        self.is_trust_scoped = is_trust_scoped
        self.redelivered = redelivered
        self.expires_at = expires_at
        self.trust_id = trust_id
        self.is_target = is_target
        self.root_execution_id = root_execution_id

        # We still use Mistral thread local variable. Maybe could consider
        # using the variable provided by oslo_context in future.
        super(MistralContext, self).__init__(overwrite=False, **kwargs)

    def get_logging_values(self):
        ctx_dict = super(MistralContext, self).get_logging_values()
        ctx_dict.update(
            {
                'root_execution_id': self.root_execution_id,
            }
        )

        return ctx_dict

    def update_store(self):
        super(MistralContext, self).update_store()

    def to_dict(self):
        """Return a dictionary of context attributes."""
        ctx_dict = super(MistralContext, self).to_dict()
        ctx_dict.update(
            {
                'user_name': self.user_name,
                'project_name': self.project_name,
                'domain_name': self.domain_name,
                'user_domain_name': self.user_domain_name,
                'project_domain_name': self.project_domain_name,
                'auth_uri': self.auth_uri,
                'auth_cacert': self.auth_cacert,
                'insecure': self.insecure,
                'service_catalog': self.service_catalog,
                'region_name': self.region_name,
                'is_trust_scoped': self.is_trust_scoped,
                'redelivered': self.redelivered,
                'expires_at': self.expires_at,
                'trust_id': self.trust_id,
                'is_target': self.is_target,
                'root_execution_id': self.root_execution_id,
            }
        )

        return ctx_dict

    @classmethod
    def from_dict(cls, values, **kwargs):
        """Construct a context object from a provided dictionary."""
        kwargs.setdefault('auth_uri', values.get('auth_uri'))
        kwargs.setdefault('auth_cacert', values.get('auth_cacert'))
        kwargs.setdefault('insecure', values.get('insecure', False))
        kwargs.setdefault('service_catalog', values.get('service_catalog'))
        kwargs.setdefault('region_name', values.get('region_name'))
        kwargs.setdefault(
            'is_trust_scoped', values.get('is_trust_scoped', False)
        )
        kwargs.setdefault('redelivered', values.get('redelivered', False))
        kwargs.setdefault('expires_at', values.get('expires_at'))
        kwargs.setdefault('trust_id', values.get('trust_id'))
        kwargs.setdefault('is_target', values.get('is_target', False))
        kwargs.setdefault('root_execution_id', values.get('root_execution_id'))

        return super(MistralContext, cls).from_dict(values, **kwargs)

    @classmethod
    def from_environ(cls, headers, env):
        kwargs = _extract_mistral_auth_params(headers)

        token_info = env.get('keystone.token_info', {})
        if not kwargs['is_target']:
            kwargs['service_catalog'] = token_info.get('token', {})
        kwargs['expires_at'] = (token_info['token']['expires_at']
                                if token_info else None)

        context = super(MistralContext, cls).from_environ(env, **kwargs)
        context.is_admin = True if 'admin' in context.roles else False

        return context


def has_ctx():
    return utils.has_thread_local(_CTX_THREAD_LOCAL_NAME)


def ctx(root_execution_id=None):
    if not has_ctx():
        raise exc.ApplicationContextNotFoundException()

    context = utils.get_thread_local(_CTX_THREAD_LOCAL_NAME)

    if root_execution_id:
        context.root_execution_id = root_execution_id
        set_ctx(context)

    return context


def set_ctx(new_ctx):
    if new_ctx:
        new_ctx.update_store()
    utils.set_thread_local(_CTX_THREAD_LOCAL_NAME, new_ctx)


def _extract_mistral_auth_params(headers):
    service_catalog = None

    if headers.get("X-Target-Auth-Uri"):
        insecure_header = headers.get('X-Target-Insecure', 'False')
        if insecure_header == 'False':
            insecure = False
        elif insecure_header == 'True':
            insecure = True
        else:
            raise (exc.MistralException(
                'X-Target-Insecure must be either "True", "False" or not '
                'provided. The default is "False".'))

        params = {
            # TODO(akovi): Target cert not handled yet
            'auth_cacert': None,
            'insecure': insecure,
            'auth_token': headers.get('X-Target-Auth-Token'),
            'auth_uri': headers.get('X-Target-Auth-Uri'),
            'project_id': headers.get('X-Target-Project-Id'),
            'user': headers.get('X-Target-User-Id'),
            'user_name': headers.get('X-Target-User-Name'),
            'region_name': headers.get('X-Target-Region-Name'),
            'is_target': True
        }
        if not params['auth_token']:
            raise (exc.MistralException(
                'Target auth URI (X-Target-Auth-Uri) target auth token '
                '(X-Target-Auth-Token) must be present'))

        # It's possible that target service catalog is not provided, in this
        # case, Mistral needs to get target service catalog dynamically when
        # talking to target openstack deployment later on.
        service_catalog = _extract_service_catalog_from_headers(
            headers
        )
    else:
        params = {
            'auth_uri': CONF.keystone_authtoken.www_authenticate_uri,
            'auth_cacert': CONF.keystone_authtoken.cafile,
            'insecure': False,
            'region_name': headers.get('X-Region-Name'),
            'is_target': False
        }

    params['service_catalog'] = service_catalog

    return params


def _extract_service_catalog_from_headers(headers):
    target_service_catalog_header = headers.get(
        'X-Target-Service-Catalog')
    if target_service_catalog_header:
        decoded_catalog = base64.b64decode(
            target_service_catalog_header).decode()
        return jsonutils.loads(decoded_catalog)
    else:
        return None


class RpcContextSerializer(messaging.Serializer):
    def __init__(self, entity_serializer=None):
        self.entity_serializer = (
            entity_serializer or serialization.get_polymorphic_serializer()
        )

    def serialize_entity(self, context, entity):
        if not self.entity_serializer:
            return entity

        return self.entity_serializer.serialize(entity)

    def deserialize_entity(self, context, entity):
        if not self.entity_serializer:
            return entity

        return self.entity_serializer.deserialize(entity)

    def serialize_context(self, context):
        ctx = context.to_dict()

        pfr = profiler.get()

        if pfr:
            ctx['trace_info'] = {
                "hmac_key": pfr.hmac_key,
                "base_id": pfr.get_base_id(),
                "parent_id": pfr.get_id()
            }

        return ctx

    def deserialize_context(self, context):
        trace_info = context.pop('trace_info', None)

        if trace_info:
            profiler.init(**trace_info)

        ctx = MistralContext.from_dict(context)

        set_ctx(ctx)

        return ctx


class AuthHook(hooks.PecanHook):
    def before(self, state):
        if state.request.path in ALLOWED_WITHOUT_AUTH:
            return

        if not CONF.pecan.auth_enable:
            return

        try:
            auth_handler = auth.get_auth_handler()
            auth_handler.authenticate(state.request)
        except Exception as e:
            msg = "Failed to validate access token: %s" % str(e)

            LOG.exception(msg)

            pecan.abort(
                status_code=401,
                detail=msg,
                headers={'Server-Error-Message': msg, "WWW-Authenticate": msg}
            )


class ContextHook(hooks.PecanHook):
    def before(self, state):
        context = MistralContext.from_environ(
            state.request.headers, state.request.environ
        )

        set_ctx(context)

    def after(self, state):
        set_ctx(None)


def create_action_context(execution_ctx):

    context = ctx()

    security_ctx = lib_ctx.SecurityContext(
        auth_cacert=context.auth_cacert,
        auth_token=context.auth_token,
        auth_uri=context.auth_uri,
        expires_at=context.expires_at,
        insecure=context.insecure,
        is_target=context.is_target,
        is_trust_scoped=context.is_trust_scoped,
        project_id=context.project_id,
        project_name=context.project_name,
        user_name=context.user_name,
        redelivered=context.redelivered,
        region_name=context.region_name,
        service_catalog=context.service_catalog,
        trust_id=context.trust_id,
    )

    ex_ctx = lib_ctx.ExecutionContext(**execution_ctx)

    return lib_ctx.ActionContext(security_ctx, ex_ctx)



# Copyright 2023 - NetCracker Technology Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import copy
import hashlib

from oslo_config import cfg


VERSIONS_KEY = "__versions"


def clear_versions(ctx):
    if VERSIONS_KEY in ctx:
        del ctx[VERSIONS_KEY]


def get_in_context_with_versions(task_ex):
    in_context = task_ex.in_context if task_ex.in_context else {}

    if not cfg.CONF.context_versioning.enabled:
        return in_context

    in_context = (copy.deepcopy(dict(in_context)))

    task_published = task_ex.published or []

    in_context.setdefault(VERSIONS_KEY, {})

    updated_keys = _get_updated_keys(task_published)

    for updated in updated_keys:
        in_context[VERSIONS_KEY].setdefault(updated, 0)
        in_context[VERSIONS_KEY][updated] += 1

    return in_context


def _get_updated_keys(published):
    updated_keys = []
    _get_published_keys_recursively(updated_keys, published)
    return updated_keys


def _get_published_keys_recursively(updated_keys, published, prefix=None):
    for key in published:
        new_prefix = key if not prefix else prefix + "." + key

        if not isinstance(published[key], dict):
            if cfg.CONF.context_versioning.hash_version_keys:
                new_prefix = hashlib.md5(
                    new_prefix.encode("utf-8")
                ).hexdigest()

            updated_keys.append(new_prefix)
        else:
            _get_published_keys_recursively(
                updated_keys,
                published[key],
                new_prefix
            )


def merge_context_by_version(ctx_left, ctx_right):
    versions_left = copy.copy(ctx_left[VERSIONS_KEY])
    versions_right = copy.copy(ctx_right[VERSIONS_KEY])

    _remove_internal_data_from_context(ctx_left)
    _remove_internal_data_from_context(ctx_right)

    result = _merge_ctx(ctx_left, versions_left, ctx_right, versions_right)

    result[VERSIONS_KEY] = _merge_versions(versions_left, versions_right)

    return result


def _get_version(key, versions):
    if key not in versions:
        return 0

    return versions[key]


def _merge_ctx(ctx_left, ver_left, ctx_right, ver_right, prefix=None):
    if ctx_left is None:
        return ctx_right

    if ctx_right is None:
        return ctx_left

    for k, v in ctx_right.items():
        if k not in ctx_left:
            ctx_left[k] = v
        else:
            left_v = ctx_left[k]

            new_prefix = k if not prefix else prefix + "." + k

            if isinstance(left_v, dict) and isinstance(v, dict):
                _merge_ctx(left_v, ver_left, v, ver_right, new_prefix)
            else:
                if cfg.CONF.context_versioning.hash_version_keys:
                    new_prefix = hashlib.md5(
                        new_prefix.encode("utf-8")
                    ).hexdigest()
                l_ver = _get_version(new_prefix, ver_left)
                r_ver = _get_version(new_prefix, ver_right)
                if r_ver > l_ver:
                    ctx_left[k] = v

    return ctx_left


def _merge_versions(ver_left, ver_right):
    for key in ver_right:
        if key not in ver_left:
            ver_left[key] = ver_right[key]
        ver_left[key] = max(ver_left[key], ver_right[key])

    return ver_left


def _remove_internal_data_from_context(ctx):
    if '__task_execution' in ctx:
        del ctx['__task_execution']
    del ctx[VERSIONS_KEY]



# Copyright (c) 2016 Intel Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from kombu import mixins as mx
from unittest import mock


# Hack for making tests works with kombu listener
mixins = mx

producer = mock.MagicMock()

producers = mock.MagicMock()
producers.__getitem__ = lambda *args, **kwargs: producer

connection = mock.MagicMock()

connections = mock.MagicMock()
connections.__getitem__ = lambda *args, **kwargs: connection

serialization = mock.MagicMock()


def BrokerConnection(*args, **kwargs):
    return mock.MagicMock()


def Exchange(*args, **kwargs):
    return mock.MagicMock()


def Queue(*args, **kwargs):
    return mock.MagicMock()


def Consumer(*args, **kwargs):
    return mock.MagicMock()



# Copyright 2015 - Mirantis, Inc.
# Copyright 2015 - StackStorm, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import copy
import json
import jsonschema
from osprofiler import profiler
import re

from mistral import exceptions as exc
from mistral import expressions as expr
from mistral.expressions.jinja_expression import ANY_JINJA_REGEXP
from mistral.expressions.yaql_expression import INLINE_YAQL_REGEXP
from mistral.lang import types
from mistral_lib import utils

ACTION_PATTERNS = {
    "command": r"[\w\.]+[^=\(\s\"]*",
    "yaql_expression": INLINE_YAQL_REGEXP,
    "jinja_expression": ANY_JINJA_REGEXP,
}
CMD_PTRN = re.compile(
    "^({})".format("|".join(ACTION_PATTERNS.values()))
)

EXPRESSION = '|'.join([expr.patterns[name] for name in expr.patterns])
_ALL_IN_BRACKETS = r"\[.*\]\s*"
_ALL_IN_QUOTES = r"\"[^\"]*\"\s*"
_ALL_IN_APOSTROPHES = r"'[^']*'\s*"
_DIGITS = r"\d+"
_TRUE = "true"
_FALSE = "false"
_NULL = "null"

ALL = (
    _ALL_IN_QUOTES, _ALL_IN_APOSTROPHES, EXPRESSION,
    _ALL_IN_BRACKETS, _TRUE, _FALSE, _NULL, _DIGITS
)

PARAMS_PTRN = re.compile(r"([-_\w]+)=(%s)" % "|".join(ALL))

# {(base_spec_cls, polymorphic_value): spec_cls}
_POLYMORPHIC_CACHE = {}


@profiler.trace('lang-base-instantiate-spec', hide_args=True)
def instantiate_spec(spec_cls, data, validate=False):
    """Instantiates specification accounting for specification hierarchies.

    :param spec_cls: Specification concrete or base class. In case if base
        class or the hierarchy is provided this method relies on attributes
        _polymorphic_key and _polymorphic_value in order to find a concrete
        class that needs to be instantiated.
    :param data: Raw specification data as a dictionary.
    :type data: dict
    :param validate: If it's False then semantics and schema validation
        will be skipped.
    :type validate: bool
    """

    if issubclass(spec_cls, BaseSpecList):
        # Ignore polymorphic search for specification lists because
        # it doesn't make sense for them.
        return spec_cls(data, validate)

    if not hasattr(spec_cls, '_polymorphic_key'):
        spec = spec_cls(data, validate)

        if validate:
            spec.validate_semantics()

        return spec

    # In order to do polymorphic search we need to make sure that
    # a spec is backed by a dictionary. Otherwise we can't extract
    # a polymorphic key.
    if not isinstance(data, dict):
        raise exc.InvalidModelException(
            "A specification with polymorphic key must be backed by"
            " a dictionary [spec_cls=%s, data=%s]" % (spec_cls, data)
        )

    key = spec_cls._polymorphic_key

    if not isinstance(key, tuple):
        key_name = key
        key_default = None
    else:
        key_name = key[0]
        key_default = key[1]

    polymorphic_val = data.get(key_name, key_default)

    global _POLYMORPHIC_CACHE

    cache_key = (spec_cls, polymorphic_val)

    concrete_spec_cls = _POLYMORPHIC_CACHE.get(cache_key)

    if concrete_spec_cls is None:
        for cls in utils.iter_subclasses(spec_cls):
            if not hasattr(cls, '_polymorphic_value'):
                raise exc.DSLParsingException(
                    "Class '%s' is expected to have attribute"
                    " '_polymorphic_value' because it's a part of"
                    " specification hierarchy inherited "
                    "from class '%s'." % (cls, spec_cls)
                )

            if cls._polymorphic_value == polymorphic_val:
                concrete_spec_cls = cls

                _POLYMORPHIC_CACHE[cache_key] = concrete_spec_cls

    if concrete_spec_cls is None:
        raise exc.DSLParsingException(
            'Failed to find a specification class to instantiate '
            '[spec_cls=%s, data=%s]' % (spec_cls, data)
        )

    spec = concrete_spec_cls(data, validate)

    if validate:
        spec.validate_semantics()

    return spec


class BaseSpec(object):
    """Base class for all DSL specifications.

    It represents a DSL entity such as workflow or task as a python object
    providing more convenient API to analyse DSL than just working with raw
    data in form of a dictionary. Specification classes also implement
    all required validation logic by overriding instance methods
    'validate_schema()' and 'validate_semantics()'.

    Note that the specification mechanism allows to have polymorphic entities
    in DSL. For example, if we find it more convenient to have separate
    specification classes for different types of workflow (i.e. 'direct' and
    'reverse') we can do so. In this case, in order to instantiate them
    correctly method 'instantiate_spec' must always be used where argument
    'spec_cls' must be a root class of the specification hierarchy containing
    class attribute '_polymorhpic_key' pointing to a key in raw data relying
    on which we can find a concrete class. Concrete classes then must all have
    attribute '_polymorhpic_value' corresponding to a value in a raw data.
    Attribute '_polymorhpic_key' can be either a string or a tuple of size two
    where the first value is a key name itself and the second value is a
    default polymorphic value that must be used if raw data doesn't contain
    a configured key at all. An example of this situation is when we don't
    specify a workflow type in DSL. In this case, we assume it's 'direct'.
    """

    # See http://json-schema.org
    _schema = {'type': 'object'}

    _meta_schema = {'type': 'object'}

    _full_schema = None

    _definitions = {}

    _version = '2.0'

    @classmethod
    def get_schema(cls, includes=('meta', 'definitions')):
        if cls._full_schema is not None:
            return cls._full_schema

        schema = copy.deepcopy(cls._schema)

        schema['properties'] = utils.merge_dicts(
            schema.get('properties', {}),
            cls._meta_schema.get('properties', {}),
            overwrite=False
        )

        if includes and 'meta' in includes:
            schema['required'] = list(
                set(schema.get('required', []) +
                    cls._meta_schema.get('required', []))
            )

        if includes and 'definitions' in includes:
            schema['definitions'] = utils.merge_dicts(
                schema.get('definitions', {}),
                cls._definitions,
                overwrite=False
            )

        cls._full_schema = schema

        return schema

    def __init__(self, data, validate):
        self._data = data
        self._validate = validate

        if validate:
            self.validate_schema()

    @profiler.trace('lang-base-spec-validate-schema', hide_args=True)
    def validate_schema(self):
        """Validates DSL entity schema that this specification represents.

        By default, this method just validate schema of DSL entity that this
        specification represents using "_schema" class attribute.
        Additionally, child classes may implement additional logic to validate
        more specific things like YAQL expressions in their fields.

        Note that this method is called before construction of specification
        fields and validation logic should only rely on raw data provided as
        a dictionary accessible through '_data' instance field.
        """

        try:
            jsonschema.validate(self._data, self.get_schema())
        except jsonschema.ValidationError as e:
            raise exc.InvalidModelException("Invalid DSL: %s" % e)

    def validate_semantics(self):
        """Validates semantics of specification object.

        Child classes may implement validation logic to check things like
        integrity of corresponding data structure (e.g. task graph) or
        other things that can't be expressed in JSON schema.

        This method is called after specification has been built (i.e.
        its initializer has finished it's work) so that validation logic
        can rely on initialized specification fields.
        """
        pass

    def validate_expr(self, dsl_part):
        if isinstance(dsl_part, str):
            expr.validate(dsl_part)
        elif isinstance(dsl_part, (list, tuple)):
            for expression in dsl_part:
                if isinstance(expression, str):
                    expr.validate(expression)
        elif isinstance(dsl_part, dict):
            for expression in dsl_part.values():
                if isinstance(expression, str):
                    expr.validate(expression)

    def _spec_property(self, prop_name, spec_cls):
        prop_val = self._data.get(prop_name)

        if prop_val is None:
            return None

        return instantiate_spec(spec_cls, prop_val, self._validate)

    def _group_spec(self, spec_cls, *prop_names):
        if not prop_names:
            return None

        data = {}

        for prop_name in prop_names:
            prop_val = self._data.get(prop_name)

            if prop_val:
                data[prop_name] = prop_val

        return instantiate_spec(spec_cls, data, self._validate)

    def _inject_version(self, prop_names):
        for prop_name in prop_names:
            prop_data = self._data.get(prop_name)

            if isinstance(prop_data, dict):
                prop_data['version'] = self._version

    def _as_dict(self, prop_name):
        prop_val = self._data.get(prop_name)

        if not prop_val:
            return {}

        if isinstance(prop_val, dict):
            return prop_val
        elif isinstance(prop_val, list):
            result = {}

            for t in prop_val:
                result.update(t if isinstance(t, dict) else {t: ''})

            return result
        elif isinstance(prop_val, str):
            return {prop_val: ''}

    @staticmethod
    @profiler.trace('lang-base-parse-cmd-and-input', hide_args=True)
    def _parse_cmd_and_input(cmd_str):
        if ' ' not in cmd_str:
            return cmd_str, {}

        # TODO(rakhmerov): Try to find a way with one expression.
        cmd_matcher = CMD_PTRN.search(cmd_str)

        if not cmd_matcher:
            msg = "Invalid action/workflow task property: %s" % cmd_str

            raise exc.InvalidModelException(msg)

        cmd = cmd_matcher.group()

        params = {}

        for match in re.findall(PARAMS_PTRN, cmd_str):
            k = match[0]
            # Remove embracing quotes.
            v = match[1].strip()
            if v[0] == '"' or v[0] == "'":
                v = v[1:-1]
            else:
                try:
                    v = json.loads(v)
                except Exception:
                    pass

            params[k] = v

        return cmd, params

    def to_dict(self):
        return self._data

    def get_version(self):
        return self._version

    def __repr__(self):
        return "%s %s" % (self.__class__.__name__, self.to_dict())


class BaseListSpec(BaseSpec):
    item_class = None

    _schema = {
        "type": "object",
        "properties": {
            "version": types.VERSION
        },
        "additionalProperties": types.NONEMPTY_DICT,
        "required": ["version"],
    }

    def __init__(self, data, validate):
        super(BaseListSpec, self).__init__(data, validate)

        self.items = []

        for k, v in data.items():
            if k != 'version':
                v['name'] = k
                self._inject_version([k])

                self.items.append(
                    instantiate_spec(self.item_class, v, validate)
                )

    def validate_schema(self):
        super(BaseListSpec, self).validate_schema()

        if len(self._data.keys()) < 2:
            raise exc.InvalidModelException(
                'At least one item must be in the list [data=%s].' %
                self._data
            )

    def get_items(self):
        return self.items

    def __getitem__(self, idx):
        return self.items[idx]

    def __len__(self):
        return len(self.items)


class BaseSpecList(object):
    item_class = None

    _version = '2.0'

    def __init__(self, data, validate):
        self.items = {}

        for k, v in data.items():
            if k != 'version':
                # At this point, we don't know if item schema is valid,
                # it may not be even a dictionary. So we should check the
                # type first before manipulating with it.
                if isinstance(v, dict):
                    v['name'] = k
                    v['version'] = self._version

                self.items[k] = instantiate_spec(self.item_class, v, validate)

    def item_keys(self):
        return self.items.keys()

    def __iter__(self):
        return iter(self.items.values())

    def __getitem__(self, name):
        return self.items.get(name)

    def __len__(self):
        return len(self.items)

    def get(self, name):
        return self.__getitem__(name)



# Copyright 2015 - Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock

from oslo_config import cfg
from oslo_utils import importutils
import testtools

from mistral.db.v2 import api as db_api
from mistral.expressions import yaql_expression
from mistral.services import workflows as wf_service
from mistral.tests.unit.engine import base
from mistral.utils import javascript
from mistral.workflow import states

# Use the set_default method to set value otherwise in certain test cases
# the change in value is not permanent.
cfg.CONF.set_default('auth_enable', False, group='pecan')


JAVASCRIPT_WORKFLOW = """
version: "2.0"
wf:
  input:
    - length
  tasks:
    task1:
      action: std.javascript
      input:
        script: |
          let numberSequence = Array.from({length: $['length']},
                                          (x, i) => i);
          let evenNumbers = numberSequence.filter(x => x % 2 === 0);

          return evenNumbers.length;
        context: <% $ %>
      publish:
        res: <% task().result %>
"""


def fake_evaluate(_, context):
    return context['length'] / 2


class JavaScriptEngineTest(base.EngineTestCase):

    @testtools.skipIf(not importutils.try_import('py_mini_racer'),
                      'This test requires that py_mini_racer library was '
                      'installed')
    def test_py_mini_racer_javascript_action(self):
        cfg.CONF.set_default('js_implementation', 'py_mini_racer')

        length = 1000

        wf_service.create_workflows(JAVASCRIPT_WORKFLOW)

        # Start workflow.
        wf_ex = self.engine.start_workflow(
            'wf',
            wf_input={'length': length}
        )

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            task_ex = wf_ex.task_executions[0]

        self.assertEqual(states.SUCCESS, task_ex.state)
        self.assertDictEqual({}, task_ex.runtime_context)

        self.assertEqual(length / 2, task_ex.published['res'])

    @testtools.skipIf(not importutils.try_import('py_mini_racer'),
                      'This test requires that py_mini_racer library was '
                      'installed')
    def test_py_mini_racer_javascript_action_disabled_yaql_conversion(self):
        cfg.CONF.set_default('js_implementation', 'py_mini_racer')

        # Both input and output data conversion in YAQL need to be disabled
        # so that we're sure that there won't be any surprises from YAQL
        # like some YAQL internal types included in expression results.
        self.override_config('convert_input_data', False, 'yaql')
        self.override_config('convert_output_data', False, 'yaql')

        # Setting YAQL engine to None so it reinitialized again with the
        # right values upon the next use.
        yaql_expression.YAQL_ENGINE = None

        wf_text = """---
        version: '2.0'

        wf:
          input:
            - param: default_val

          tasks:
            task1:
              action: std.js
              input:
                context: <% $ %>
                script: >
                  return $.param
              publish:
                result: <% task().result %>
        """

        wf_service.create_workflows(wf_text)

        # Start workflow.
        wf_ex = self.engine.start_workflow('wf')

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            # Note: We need to reread execution to access related tasks.
            wf_ex = db_api.get_workflow_execution(wf_ex.id)

            t_ex = self._assert_single_item(
                wf_ex.task_executions,
                name='task1'
            )

            self.assertDictEqual({'result': 'default_val'}, t_ex.published)

    @mock.patch.object(javascript, 'evaluate', fake_evaluate)
    def test_fake_javascript_action_data_context(self):
        length = 1000

        wf_service.create_workflows(JAVASCRIPT_WORKFLOW)

        # Start workflow.
        wf_ex = self.engine.start_workflow(
            'wf',
            wf_input={'length': length}
        )

        self.await_workflow_success(wf_ex.id)

        with db_api.transaction():
            wf_ex = db_api.get_workflow_execution(wf_ex.id)
            task_ex = wf_ex.task_executions[0]

        self.assertEqual(states.SUCCESS, task_ex.state)
        self.assertDictEqual({}, task_ex.runtime_context)

        self.assertEqual(length / 2, task_ex.published['res'])
